from __future__ import annotations

import json
import pathlib
from typing import Any

import mitzu.model as M
import pandas as pd
from mitzu.adapters.sqlite_adapter import SQLiteAdapter

VALUE_SEPARATOR = "###"


class FileAdapter(SQLiteAdapter):
    def __init__(self, project: M.Project):
        super().__init__(project)

    def get_engine(self) -> Any:
        if self._engine is None:
            self._engine = super().get_engine()
            for ed_table in self.project.event_data_tables:
                df = self._read_file(ed_table)
                df.to_sql(
                    name=ed_table.table_name,
                    con=self._engine,
                    index=False,
                )
        return self._engine

    def _read_file(self, event_data_table: M.EventDataTable) -> pd.DataFrame:
        project = self.project
        extension: str = project.connection.extra_configs["file_type"]
        path = project.connection.extra_configs["path"]
        table_name = event_data_table.table_name
        file_loc = pathlib.Path(path, f"{table_name}.{extension}")
        if extension.endswith("csv"):
            df = pd.read_csv(file_loc, header=0)
        elif extension.endswith("json"):
            df = pd.read_json(file_loc)
        elif extension.endswith("parquet"):
            df = pd.read_parquet(
                file_loc,
            )
        else:
            raise Exception("Extension not supported: " + extension)
        df[event_data_table.event_time_field._get_name()] = pd.to_datetime(
            df[event_data_table.event_time_field._get_name()]
        )
        return self._fix_complex_types(df)

    def _fix_complex_types(self, df: pd.DataFrame) -> pd.DataFrame:
        for col in df.columns:
            obj = df[col][0]
            if pd.api.types.is_dict_like(obj):
                df[col] = df[col].apply(lambda val: json.dumps(val, default=str))
            elif pd.api.types.is_list_like(obj):
                if type(obj) == tuple:
                    df[col] = df[col].apply(
                        lambda val: json.dumps(dict(val), default=str)
                    )
                else:
                    df[col] = df[col].apply(lambda val: json.dumps(val, default=str))
        return df



from uuid import uuid4
import os
from typing import Tuple, Optional

# base
LAUNCH_UID = str(uuid4())
HEALTH_CHECK_PATH = os.getenv("HEALTH_CHECK_PATH", "/_health")


# dash
GRAPH_POLL_INTERVAL_MS = int(os.getenv("GRAPH_POLL_INTERVAL_MS", 100))
BACKGROUND_CALLBACK = bool(os.getenv("BACKGROUND_CALLBACK", "true").lower() != "false")
DASH_ASSETS_FOLDER = os.getenv("DASH_ASSETS_FOLDER", "assets")
DASH_ASSETS_URL_PATH = os.getenv("DASH_ASSETS_URL_PATH", "assets")
DASH_SERVE_LOCALLY = bool(os.getenv("DASH_SERVE_LOCALLY", True))
DASH_TITLE = os.getenv("DASH_TITLE", "Mitzu")
DASH_FAVICON_PATH = os.getenv("DASH_FAVICON_PATH", "assets/favicon.ico")
DASH_COMPRESS_RESPONSES = bool(os.getenv("DASH_COMPRESS_RESPONSES", True))
DASH_LOGO_PATH = os.getenv("DASH_LOGO_PATH", "/assets/mitzu-logo-light.svg")


# auth
AUTH_BACKEND = os.getenv("AUTH_BACKEND")
AUTH_ALLOWED_EMAIL_DOMAIN = os.getenv("AUTH_ALLOWED_EMAIL_DOMAIN")
AUTH_JWT_SECRET = os.getenv("AUTH_JWT_SECRET", "mitzu-dev-env")
AUTH_SESSION_TIMEOUT = os.getenv("AUTH_SESSION_TIMEOUT", 7 * 24 * 60 * 60)
AUTH_ROOT_PASSWORD = os.getenv("AUTH_ROOT_PASSWORD", "test")
AUTH_ROOT_USER_EMAIL = os.getenv("AUTH_ROOT_USER_EMAIL", "root@local")


# cache
CACHE_EXPIRATION = int(os.getenv("CACHE_EXPIRATION", "600"))

# redis cache
QUEUE_REDIS_HOST = os.getenv("QUEUE_REDIS_HOST")
QUEUE_REDIS_PORT = int(os.getenv("QUEUE_REDIS_PORT", 6379))
QUEUE_REDIS_DB = os.getenv("STORAGE_REDIS_DB")
QUEUE_REDIS_USERNAME = os.getenv("QUEUE_REDIS_USERNAME")
QUEUE_REDIS_PASSWORD = os.getenv("QUEUE_REDIS_PASSWORD")

STORAGE_REDIS_HOST = os.getenv("STORAGE_REDIS_HOST")
STORAGE_REDIS_PORT = int(os.getenv("STORAGE_REDIS_PORT", 6379))
STORAGE_REDIS_DB = os.getenv("STORAGE_REDIS_DB")
STORAGE_REDIS_USERNAME = os.getenv("STORAGE_REDIS_USERNAME")
STORAGE_REDIS_PASSWORD = os.getenv("STORAGE_REDIS_PASSWORD")


# disk cache
DISK_CACHE_PATH = os.getenv("DISK_CACHE_PATH", "./cache")

# storage
SETUP_SAMPLE_PROJECT = bool(
    os.getenv("SETUP_SAMPLE_PROJECT", "true").lower() != "false"
)


KALEIDO_CONFIGS = os.getenv("KALEIDO_CONFIGS", "--disable-gpu-*,--single-process")


def get_kaleido_configs() -> Optional[Tuple[str, ...]]:
    if KALEIDO_CONFIGS:
        return tuple(KALEIDO_CONFIGS.split(","))
    else:
        return None



from __future__ import annotations

import dash.development.base_component as bc

import dash_bootstrap_components as dbc
from dash import html
from typing import List, Optional, cast
import mitzu.webapp.storage as S
import mitzu.webapp.pages.paths as P
import flask
import mitzu.webapp.dependencies as DEPS


OFF_CANVAS_TOGGLER = "off-canvas-toggler"


def create_explore_button_col(
    storage: Optional[S.MitzuStorage] = None, project_name: Optional[str] = None
):
    if storage is None:
        storage = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        ).storage

    project_ids = storage.list_projects()
    projects = [storage.get_project(p_id) for p_id in project_ids]
    return dbc.Col(
        dbc.DropdownMenu(
            children=[
                dbc.DropdownMenuItem(
                    children=p.project_name,
                    href=P.create_path(P.PROJECTS_EXPLORE_PATH, project_id=p.id),
                )
                for p in projects
            ],
            size="sm",
            color="light",
            label="explore" if project_name is None else project_name,
            class_name="d-inline-block",
        ),
        width="auto",
    )


def create_mitzu_navbar(
    id: str,
    children: List[bc.Component] = [],
    storage: Optional[S.MitzuStorage] = None,
    create_explore_button: bool = True,
    off_canvas_toggler_visible: bool = True,
    project_name: Optional[str] = None,
) -> dbc.Navbar:

    navbar_children = [
        dbc.Col(
            dbc.Button(
                html.B(className="bi bi-list"),
                color="primary",
                size="sm",
                className="me-3",
                id={"type": OFF_CANVAS_TOGGLER, "index": id},
                style={
                    "display": "inline-block" if off_canvas_toggler_visible else "none"
                },
            ),
            width="auto",
        ),
    ]
    if create_explore_button:
        navbar_children.append(create_explore_button_col(storage, project_name))

    navbar_children.extend([dbc.Col(comp) for comp in children])
    res = dbc.Navbar(
        dbc.Container(
            [
                dbc.Row(
                    children=navbar_children,
                    className="g-2",
                ),
            ],
            fluid=True,
        ),
        class_name="mb-3",
        color="dark",
    )

    return res



from dataclasses import dataclass
import mitzu.webapp.storage as S
from typing import Callable, Dict, List, Optional
import mitzu.model as M


class EventServiceException(Exception):
    pass


@dataclass
class EventsService:

    storage: S.MitzuStorage

    def list_all_projects(self) -> List[M.Project]:
        project_ids = self.storage.list_projects()
        return [
            self.storage.get_project(pid) for pid in project_ids
        ]  # todo move this logic to storage

    def get_project_definition(
        self, project_id: str
    ) -> Dict[M.EventDataTable, Dict[str, M.Reference[M.EventDef]]]:
        dp = self.storage.get_project(project_id)._discovered_project.get_value()
        if dp is None:
            raise EventServiceException(f"Missing discovered project for {project_id}")
        return dp.definitions

    def discover_project(
        self,
        project_id: str,
        callback: Callable[
            [
                M.EventDataTable,
                Dict[str, M.Reference[M.EventDef]],
                Optional[Exception],
                int,
                int,
            ],
            None,
        ],
    ) -> Dict[M.EventDataTable, Dict[str, M.Reference[M.EventDef]]]:
        project = self.storage.get_project(project_id)
        edt_count = len(project.event_data_tables)
        edts = []

        def edt_callback(
            edt: M.EventDataTable,
            defs: Dict[str, M.Reference[M.EventDef]],
            exc: Optional[Exception],
        ):
            if exc is None:
                self.storage.set_event_data_table_definition(
                    project_id, edt_full_name=edt.get_full_name(), definitions=defs
                )
            edts.append(edt)
            callback(edt, defs, exc, len(edts), edt_count)

        discovered_project = project.discover_project(False, edt_callback)
        self.storage.set_project(
            project_id=discovered_project.project.id,
            project=discovered_project.project,
        )

        return discovered_project.definitions



from __future__ import annotations

from enum import Enum, auto
from typing import Optional

import dash.development.base_component as bc
import mitzu.model as M
import dash_mantine_components as dmc


METRIC_TYPE_DROPDOWN = "metric-type-dropdown"


class MetricType(Enum):
    SEGMENTATION = auto()
    CONVERSION = auto()
    RETENTION = auto()
    JOURNEY = auto()

    @classmethod
    def from_metric(cls, metric: Optional[M.Metric]) -> MetricType:
        if isinstance(metric, M.ConversionMetric):
            return MetricType.CONVERSION
        elif isinstance(metric, M.RetentionMetric):
            return MetricType.RETENTION
        else:
            return MetricType.SEGMENTATION

    @classmethod
    def parse(cls, val_str: str) -> MetricType:
        return MetricType[val_str.upper()]


def from_metric_type(metric_type: MetricType) -> bc.Component:
    return dmc.Select(
        data=[
            {
                "label": m_type.name.upper(),
                "value": m_type.name.upper(),
                "disabled": m_type == MetricType.JOURNEY,
            }
            for m_type in MetricType
        ],
        id=METRIC_TYPE_DROPDOWN,
        clearable=False,
        value=metric_type.name.upper(),
        size="xs",
        style={"width": "140px"},
    )



from __future__ import annotations

from datetime import datetime
from typing import Optional

import mitzu.adapters.generic_adapter as GA
import pandas as pd
import json


def str_to_datetime(val: str) -> Optional[datetime]:
    if val is None:
        return None
    return datetime.fromisoformat(val)


def dataframe_str_to_datetime(pdf: pd.DataFrame, column: str) -> pd.DataFrame:
    pdf[column] = pdf[column].apply(str_to_datetime)
    return pdf


def pdf_string_json_array_to_array(pdf: pd.DataFrame):
    for col in pdf.columns:
        if col != GA.EVENT_NAME_ALIAS_COL:
            pdf[col] = pdf[col].apply(
                lambda val: json.loads(val) if type(val) == str else val
            )
    return pdf



from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple

import dash.development.base_component as bc
import mitzu.model as M
from dash import Input, MATCH, Output, callback, dcc, html
from mitzu.webapp.helper import (
    find_event_field_def,
    get_enums,
    get_property_name_label,
    WITH_VALUE_CLS,
)
import dash_mantine_components as dmc
from mitzu.webapp.auth.decorator import restricted

SIMPLE_SEGMENT = "simple_segment"
SIMPLE_SEGMENT_WITH_VALUE = "simple_segment_with_value"
PROPERTY_NAME_DROPDOWN = "property_name_dropdown"
PROPERTY_OPERATOR_DROPDOWN = "property_operator_dropdown"
PROPERTY_VALUE_INPUT = "property_value_input"


OPERATOR_MAPPING = {
    M.Operator.ANY_OF: "is",
    M.Operator.NONE_OF: "is not",
    M.Operator.GT: ">",
    M.Operator.GT_EQ: ">=",
    M.Operator.LT: "<",
    M.Operator.LT_EQ: "<=",
    M.Operator.IS_NOT_NULL: "present",
    M.Operator.IS_NULL: "missing",
    M.Operator.LIKE: "like",
    M.Operator.NOT_LIKE: "not like",
}

NULL_OPERATORS = ["present", "missing"]
MULTI_OPTION_OPERATORS = [M.Operator.ANY_OF, M.Operator.NONE_OF]
BOOL_OPERATORS = [M.Operator.IS_NOT_NULL, M.Operator.IS_NULL]
CUSTOM_VAL_PREFIX = "$EQ$_"


def create_property_dropdown(
    simple_segment: M.SimpleSegment,
    discovered_project: M.DiscoveredProject,
    simple_segment_index: int,
    type_index: str,
) -> dmc.Select:
    event_name = simple_segment._left._event_name
    field_name: Optional[str] = None
    if type(simple_segment._left) == M.EventFieldDef:
        field_name = simple_segment._left._field._get_name()

    event = discovered_project.get_event_def(event_name)
    placeholder = "+ Where" if simple_segment_index == 0 else "+ And"
    fields_names = list(event._fields.keys())
    fields_names.sort(key=lambda f: f._get_name())
    options = [
        {
            "label": get_property_name_label(f._get_name()),
            "value": f"{event_name}.{f._get_name()}",
        }
        for f in fields_names
    ]

    return dmc.Select(
        data=options,
        value=None if field_name is None else f"{event_name}.{field_name}",
        placeholder=placeholder,
        searchable=True,
        clearable=True,
        className=PROPERTY_NAME_DROPDOWN,
        id={
            "type": PROPERTY_NAME_DROPDOWN,
            "index": type_index,
        },
    )


def create_value_input(
    simple_segment: M.SimpleSegment,
    discovered_project: M.DiscoveredProject,
    type_index: str,
) -> dcc.Dropdown:
    multi = simple_segment._operator in MULTI_OPTION_OPERATORS
    value = simple_segment._right
    left = simple_segment._left

    if type(left) == M.EventFieldDef:
        path = f"{left._event_name}.{left._field._get_name()}"
        enums = get_enums(path, discovered_project)
        if value is not None:
            if type(value) in (list, tuple):
                enums = list(set([*list(value), *enums]))
            else:
                enums = [value, *enums]
    else:
        enums = []

    options = [{"label": str(e), "value": e} for e in enums]
    options.sort(key=lambda v: v["label"])

    placeholder = (", ".join([str(e) for e in enums]))[0:20] + "..."

    comp_value: Any = value
    if multi:
        if value is not None and type(value) in (list, tuple):
            comp_value = list(value)
        if value is None:
            comp_value = []

    return dmc.MultiSelect(
        data=options,
        value=comp_value,
        clearable=False,
        searchable=True,
        creatable=True,
        maxSelectedValues=(100 if multi else 1),
        placeholder=placeholder,
        className=PROPERTY_VALUE_INPUT + " border-0",
        id={
            "type": PROPERTY_VALUE_INPUT,
            "index": type_index,
        },
        style={"width": "100%"},
    )


def create_property_operator_dropdown(
    simple_segment: M.SimpleSegment, type_index: str
) -> dcc.Dropdown:
    options: List[str] = []
    if type(simple_segment._left) == M.EventFieldDef:
        data_type = simple_segment._left._field._type
        if data_type == M.DataType.BOOL:
            options = [
                OPERATOR_MAPPING[k]
                for k in [
                    M.Operator.ANY_OF,
                    M.Operator.NONE_OF,
                    M.Operator.IS_NOT_NULL,
                    M.Operator.IS_NULL,
                ]
            ]
        elif data_type == M.DataType.NUMBER:
            options = [
                OPERATOR_MAPPING[k]
                for k in [
                    M.Operator.ANY_OF,
                    M.Operator.NONE_OF,
                    M.Operator.GT,
                    M.Operator.GT_EQ,
                    M.Operator.LT,
                    M.Operator.LT_EQ,
                    M.Operator.IS_NOT_NULL,
                    M.Operator.IS_NULL,
                ]
            ]
        else:
            options = [k for k in OPERATOR_MAPPING.values()]

    return dmc.Select(
        data=options,
        value=(
            OPERATOR_MAPPING[M.Operator.ANY_OF]
            if simple_segment._operator is None
            else OPERATOR_MAPPING[simple_segment._operator]
        ),
        searchable=True,
        clearable=False,
        className=PROPERTY_OPERATOR_DROPDOWN + " border-0",
        id={
            "type": PROPERTY_OPERATOR_DROPDOWN,
            "index": type_index,
        },
    )


def fix_custom_value(val: Any, data_type: M.DataType):
    if type(val) == str:
        return data_type.from_string(val)
    return val


def collect_values(value: Any, data_type: M.DataType) -> Optional[Tuple[Any, ...]]:
    if value is None:
        return None
    if type(value) in (list, tuple):
        return tuple([fix_custom_value(v, data_type) for v in value])
    else:
        return tuple([fix_custom_value(value, data_type)])


def from_all_inputs(
    discovered_project: Optional[M.DiscoveredProject],
    simple_segment: Dict[str, Any],
) -> Optional[M.SimpleSegment]:

    if discovered_project is None:
        return None

    property_path = simple_segment.get(PROPERTY_NAME_DROPDOWN)
    if property_path is None:
        return None

    event_field_def = find_event_field_def(property_path, discovered_project)
    property_operator = simple_segment.get(PROPERTY_OPERATOR_DROPDOWN)
    if property_operator is None:
        return M.SimpleSegment(event_field_def, M.Operator.ANY_OF, None)

    if property_operator == OPERATOR_MAPPING[M.Operator.IS_NULL]:
        return M.SimpleSegment(event_field_def, M.Operator.IS_NULL, None)
    elif property_operator == OPERATOR_MAPPING[M.Operator.IS_NOT_NULL]:
        return M.SimpleSegment(event_field_def, M.Operator.IS_NOT_NULL, None)

    property_value = simple_segment.get(PROPERTY_VALUE_INPUT)
    data_type = event_field_def._field._type

    if property_operator == OPERATOR_MAPPING[M.Operator.ANY_OF]:
        return M.SimpleSegment(
            event_field_def,
            M.Operator.ANY_OF,
            collect_values(property_value, data_type),
        )
    elif property_operator == OPERATOR_MAPPING[M.Operator.NONE_OF]:
        return M.SimpleSegment(
            event_field_def,
            M.Operator.NONE_OF,
            collect_values(property_value, data_type),
        )
    else:
        if type(property_value) == list:
            if len(property_value) >= 1:
                property_value = property_value[0]
            else:
                property_value = None
        for op, op_str in OPERATOR_MAPPING.items():
            if op_str == property_operator:
                fixed_val = fix_custom_value(property_value, data_type)
                if fixed_val == []:
                    fixed_val = None
                return M.SimpleSegment(event_field_def, op, fixed_val)

        raise ValueError(f"Not supported Operator { property_operator }")


def from_simple_segment(
    simple_segment: M.SimpleSegment,
    discovered_project: M.DiscoveredProject,
    parent_type_index: str,
    simple_segment_index: int,
) -> bc.Component:
    type_index = f"{parent_type_index}-{simple_segment_index}"
    prop_dd = create_property_dropdown(
        simple_segment, discovered_project, simple_segment_index, type_index
    )
    children = [prop_dd]
    if simple_segment._operator is not None:
        operator_dd = create_property_operator_dropdown(simple_segment, type_index)
        children.append(operator_dd)
        if simple_segment._operator not in BOOL_OPERATORS:
            value_input = create_value_input(
                simple_segment, discovered_project, type_index
            )
            children.append(value_input)

    component = html.Div(
        id={"type": SIMPLE_SEGMENT, "index": type_index},
        children=children,
        className=(
            SIMPLE_SEGMENT
            + " "
            + ("" if prop_dd.value is None else WITH_VALUE_CLS)
            + " border border-0 rounded-2"
        ),
    )

    return component


@callback(
    Output(
        {
            "type": PROPERTY_NAME_DROPDOWN,
            "index": MATCH,
        },
        "style",
    ),
    Input(
        {
            "type": PROPERTY_NAME_DROPDOWN,
            "index": MATCH,
        },
        "searchValue",
    ),
    Input(
        {
            "type": PROPERTY_NAME_DROPDOWN,
            "index": MATCH,
        },
        "value",
    ),
    prevent_initial_call=True,
)
@restricted
def property_name_style_update(search_value: str, value: str):
    # If the user is searching this search value and value is not equal.
    # In that case we need to do LTR
    if value is not None:
        fixed_value = get_property_name_label(".".join(value.split(".")[1:]))
        return {"direction": "ltr" if fixed_value != search_value else "rtl"}
    else:
        return {"direction": "ltr"}



from __future__ import annotations

import flask
from typing import cast
from dash import (
    html,
    register_page,
)
import dash_bootstrap_components as dbc
import dash.development.base_component as bc
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.service.user_service as user_service
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
from mitzu.webapp.helper import TBL_CLS, TBL_HEADER_CLS
from mitzu.webapp.auth.decorator import restricted_layout
import mitzu.webapp.service.user_service as US


ADD_USER_BUTTON = "user_add_user"
USERS_TABLE_ID = "users_table"
USERS_TBL_BODY = "users_tbl_body"

register_page(__name__, path=P.USERS_PATH, title="Mitzu - Users")


def create_users_table(user_service: user_service.UserService):
    table_header = html.Thead(
        html.Tr(
            [
                html.Th("User Email", className=TBL_HEADER_CLS + " fw-bold"),
                html.Th("Role", className=TBL_HEADER_CLS + " fw-bold"),
                html.Th("", className=TBL_HEADER_CLS),
            ],
        )
    )

    sorted_users = sorted(user_service.list_users(), key=lambda u: u.email)
    table_rows = []
    for user in sorted_users:
        table_rows.append(
            html.Tr(
                [
                    html.Td(user.email, className=TBL_CLS),
                    html.Td(user.role.value, className=TBL_CLS),
                    html.Td(
                        dbc.Button(
                            "Edit",
                            color="info",
                            href=P.create_path(P.USERS_HOME_PATH, user_id=user.id),
                            size="sm",
                            className=TBL_CLS,
                        ),
                    ),
                ],
            )
        )

    table_body = html.Tbody(table_rows, id=USERS_TBL_BODY)

    return dbc.Table(
        [table_header, table_body],
        hover=False,
        responsive=True,
        striped=True,
        size="sm",
        id=USERS_TABLE_ID,
    )


@restricted_layout
def layout(**query_params) -> bc.Component:
    deps = cast(DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY))
    user_service = deps.user_service
    if user_service is None:
        raise ValueError("User service is not set")

    authorizer = deps.authorizer
    if authorizer is None:
        raise ValueError("Authorizer is not set")

    is_admin = authorizer.get_current_user_role(flask.request) == US.Role.ADMIN

    table = create_users_table(user_service)

    return html.Div(
        [
            NB.create_mitzu_navbar("users_list", []),
            dbc.Container(
                [
                    html.H4("Registered Users"),
                    html.Hr(),
                    table,
                ]
                + (
                    [
                        html.Hr(),
                        dbc.Button(
                            [html.I(className="bi bi-plus-circle me-1"), "Add user"],
                            color="primary",
                            href=P.create_path(P.USERS_HOME_PATH, user_id="new"),
                            id=ADD_USER_BUTTON,
                        ),
                    ]
                    if is_admin
                    else []
                )
            ),
        ]
    )



from __future__ import annotations

import mitzu.model as M
import mitzu.adapters.generic_adapter as GA
import pandas as pd
import mitzu.visualization.labels as L
import mitzu.visualization.common as C
import mitzu.visualization.tooltips as T
import mitzu.visualization.titles as TI
import mitzu.visualization.transform_conv as TC
import mitzu.visualization.transform_retention as TR
from mitzu.helper import value_to_label
from typing import Optional

STEP_COL = "_step"


def filter_top_groups(
    pdf: pd.DataFrame,
    metric: M.Metric,
    order_by_col: str,
) -> pd.DataFrame:
    max = metric._max_group_count
    g_users = (
        pdf[[GA.GROUP_COL, order_by_col]].groupby(GA.GROUP_COL).sum().reset_index()
    )
    if g_users.shape[0] > 0:
        g_users = g_users.sort_values(order_by_col, ascending=False)
    g_users = g_users.head(max)
    top_groups = list(g_users[GA.GROUP_COL].values)
    return pdf[pdf[GA.GROUP_COL].isin(top_groups)]


def get_color_label(metric: M.Metric):
    if metric._group_by is not None:
        return value_to_label(metric._group_by._field._get_name())
    elif (
        isinstance(metric, M.RetentionMetric)
        and metric._chart_type == M.SimpleChartType.LINE
    ):
        return "Cohort"
    return ""


def get_hover_mode(pdf: pd.DataFrame, metric: M.Metric) -> str:
    group_count = len(list(pdf[C.COLOR_COL].values))
    if metric._time_group == M.TimeGroup.TOTAL:
        if metric._group_by is None:
            return "x"
        else:
            return "closest" if group_count > 4 else "x"
    else:
        if metric._group_by is None:
            return "x"
        else:
            return "closest" if group_count > 4 else "x"


def get_default_chart_type(metric: M.Metric) -> M.SimpleChartType:
    if metric._time_group == M.TimeGroup.TOTAL:
        return M.SimpleChartType.BAR
    else:
        if isinstance(metric, M.SegmentationMetric) or isinstance(
            metric, M.ConversionMetric
        ):
            return M.SimpleChartType.LINE
        elif isinstance(metric, M.RetentionMetric):
            return M.SimpleChartType.HEATMAP
        else:
            raise ValueError(f"No default chart type defined for {type(metric)}")


def get_preprocessed_conversion_dataframe(
    pdf: pd.DataFrame, metric: M.ConversionMetric, suffix: str
) -> pd.DataFrame:
    pdf = TC.fix_conversion_na_cols(pdf, metric)
    if metric._agg_type in [
        M.AggType.AVERAGE_TIME_TO_CONV,
        M.AggType.PERCENTILE_TIME_TO_CONV,
    ]:
        pdf = TC.fix_conv_times_pdf(pdf, metric)

    pdf = filter_top_groups(pdf, metric, order_by_col=f"{GA.USER_COUNT_COL}_1")

    if metric._time_group == M.TimeGroup.TOTAL:
        pdf = TC.get_melted_conv_pdf(pdf, metric)
    else:
        funnel_length = len(metric._conversion._segments)
        pdf[GA.AGG_VALUE_COL] = pdf[f"{GA.AGG_VALUE_COL}_{funnel_length}"]
        pdf[GA.USER_COUNT_COL] = pdf[f"{GA.USER_COUNT_COL}_{funnel_length}"]

    pdf = TC.get_conversion_mapping(pdf, metric)

    pdf[C.TEXT_COL] = pdf[C.Y_AXIS_COL].apply(lambda val: (f"{val:.1f} {suffix}"))
    pdf = T.add_conversion_tooltip(pdf, metric, suffix)

    pdf = pdf.sort_values(
        [C.X_AXIS_COL, C.COLOR_COL, GA.USER_COUNT_COL], ascending=[True, True, False]
    )
    return pdf


def get_preprocessed_segmentation_dataframe(
    pdf: pd.DataFrame, metric: M.SegmentationMetric
):
    pdf[GA.GROUP_COL] = pdf[GA.GROUP_COL].fillna("n/a").astype(str).fillna("n/a")
    pdf = filter_top_groups(pdf, metric, order_by_col=GA.AGG_VALUE_COL)
    pdf = pdf.rename(
        columns={
            GA.DATETIME_COL: C.X_AXIS_COL,
            GA.AGG_VALUE_COL: C.Y_AXIS_COL,
            GA.GROUP_COL: C.COLOR_COL,
        }
    )
    pdf[C.TEXT_COL] = pdf[C.Y_AXIS_COL]
    pdf = pdf.sort_values([C.X_AXIS_COL, C.Y_AXIS_COL], ascending=[True, False])
    pdf = T.get_segmentation_tooltip(pdf, metric)

    pdf[C.COLOR_COL] = pdf[C.COLOR_COL].fillna("")
    pdf[C.X_AXIS_COL] = pdf[C.X_AXIS_COL].fillna("")

    return pdf


def get_preprocessed_retention_dataframe(
    pdf: pd.DataFrame, metric: M.RetentionMetric
) -> pd.DataFrame:
    size = pdf.shape[0]
    pdf = TR.fix_retention(pdf, metric)
    pdf = filter_top_groups(pdf, metric, order_by_col=GA.USER_COUNT_COL + "_1")
    pdf = TR.get_retention_mapping(pdf, metric)
    pdf = T.get_retention_tooltip(pdf, metric)
    pdf[C.TEXT_COL] = pdf[C.Y_AXIS_COL].apply(
        lambda val: f"{val:.1f}%" if val > 0 and size <= 200 else ""
    )
    pdf = pdf.sort_values([C.X_AXIS_COL, C.COLOR_COL], ascending=[True, True])
    return pdf


def get_simple_chart(
    metric: M.Metric, result_df: Optional[pd.DataFrame] = None
) -> C.SimpleChart:
    if metric._chart_type is None:
        chart_type = get_default_chart_type(metric)
    else:
        chart_type = metric._chart_type

    if chart_type == M.SimpleChartType.HEATMAP:
        y_axis_label = get_color_label(metric)
        color_label = L.agg_type_label(metric._agg_type, metric._agg_param)
    else:
        y_axis_label = L.agg_type_label(metric._agg_type, metric._agg_param)
        color_label = get_color_label(metric)

    if (
        chart_type not in (M.SimpleChartType.LINE, M.SimpleChartType.STACKED_AREA)
        and metric._time_group != M.TimeGroup.TOTAL
    ):
        x_axis_label_func = C.fix_date_label
    else:
        x_axis_label_func = None

    if result_df is None:
        result_df = metric.get_df()

    if isinstance(metric, M.SegmentationMetric):
        pdf = get_preprocessed_segmentation_dataframe(result_df, metric)
        return C.SimpleChart(
            x_axis_label="",
            y_axis_label=y_axis_label,
            color_label=color_label,
            hover_mode=get_hover_mode(pdf, metric),
            chart_type=chart_type,
            title=TI.get_segmentation_title(metric),
            yaxis_ticksuffix="",
            dataframe=pdf,
            x_axis_labels_func=x_axis_label_func,
        )

    if isinstance(metric, M.ConversionMetric):
        suffix = TC.get_conversion_value_suffix(result_df, metric)
        pdf = get_preprocessed_conversion_dataframe(result_df, metric, suffix)
        return C.SimpleChart(
            x_axis_label="",
            y_axis_label=y_axis_label,
            color_label=color_label,
            title=TI.get_conversion_title(metric),
            chart_type=chart_type,
            hover_mode=get_hover_mode(pdf, metric),
            yaxis_ticksuffix=suffix,
            dataframe=pdf,
            x_axis_labels_func=x_axis_label_func,
        )

    if isinstance(metric, M.RetentionMetric):
        pdf = get_preprocessed_retention_dataframe(result_df, metric)
        return C.SimpleChart(
            x_axis_label="",
            y_axis_label=y_axis_label,
            color_label=color_label,
            title=TI.get_retention_title(metric),
            chart_type=chart_type,
            hover_mode="closest",
            yaxis_ticksuffix="%",
            dataframe=pdf,
            x_axis_labels_func=C.retention_period_label,
        )

    raise Exception(f"Unsupported metric type for visualization {type(metric)}")



from dash import register_page
import dash_bootstrap_components as dbc
from dash import html
import mitzu.webapp.configs as configs
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted_layout


register_page(
    __name__,
    path=P.HOME_PATH,
    title="Mitzu",
)


@restricted_layout
def layout(**query_params):
    return html.Div(
        [
            NB.create_mitzu_navbar("home-navbar", []),
            dbc.Container(
                children=[
                    dbc.Row(
                        [
                            html.Img(
                                src=configs.DASH_LOGO_PATH,
                                height="100px",
                                className="logo",
                            )
                        ],
                        justify="center",
                    ),
                    html.Hr(),
                    dbc.Row(
                        [
                            dbc.Button(
                                [
                                    html.B(className="bi bi-play-circle me-1"),
                                    "Start exploring",
                                ],
                                color="secondary",
                                class_name="mb-3 w-25",
                                href=P.PROJECTS_PATH,
                            ),
                        ],
                        justify="center",
                    ),
                    html.Hr(),
                    dbc.Row(
                        [],
                        justify="center",
                    ),
                ]
            ),
        ]
    )



from __future__ import annotations

import mitzu.model as M
import pandas as pd
import mitzu.visualization.common as C
import mitzu.adapters.generic_adapter as GA
import mitzu.visualization.labels as L


def add_conversion_tooltip(
    pdf: pd.DataFrame, metric: M.ConversionMetric, value_suffix: str
) -> str:
    agg_type_str = L.agg_type_label(metric._agg_type, metric._agg_param)

    if metric._time_group == M.TimeGroup.TOTAL:
        pdf[C.TOOLTIP_COL] = (
            "<b>Step: </b>"
            + pdf[C.X_AXIS_COL].astype(str)
            + "<br /><b>Group: </b>"
            + pdf[C.COLOR_COL].astype(str)
            + f"<br /><br /><b>{agg_type_str}: </b>"
            + round(pdf[C.Y_AXIS_COL], 2).astype(str)
            + f" {value_suffix}"
            + "<br /><b>User count: </b>"
            + pdf[GA.USER_COUNT_COL].astype(str)
        )
    else:
        tooltip = (
            "<b>Datetime: </b>"
            + pdf[C.X_AXIS_COL].astype(str)
            + "<br /><b>Group: </b>"
            + pdf[C.COLOR_COL].astype(str)
        )
        funnel_length = len(metric._conversion._segments)
        for step in range(1, funnel_length + 1):
            tooltip += (
                f"<br /><br /><b>Step: {step}.</b>"
                + f"<br /><b>{agg_type_str}: </b>"
                + round(pdf[GA.AGG_VALUE_COL + f"_{step}"], 2).astype(str)
                + f" {value_suffix}"
                + "<br /><b>User count: </b>"
                + pdf[GA.USER_COUNT_COL + f"_{step}"].astype(str)
            )

        pdf[C.TOOLTIP_COL] = tooltip

    return pdf


def get_segmentation_tooltip(
    pdf: pd.DataFrame,
    metric: M.SegmentationMetric,
) -> pd.DataFrame:
    agg_type_str = L.agg_type_label(metric._agg_type, metric._agg_param)
    if metric._time_group == M.TimeGroup.TOTAL:
        pdf[C.TOOLTIP_COL] = (
            "<br /><b>Group: </b>"
            + pdf[C.COLOR_COL].astype(str)
            + f"<br /><br /><b>{agg_type_str}: </b>"
            + round(pdf[C.Y_AXIS_COL], 2).astype(str)
        )
    else:
        pdf[C.TOOLTIP_COL] = (
            "<b>Datetime: </b>"
            + pdf[C.X_AXIS_COL].astype(str)
            + "<br /><b>Group: </b>"
            + pdf[C.COLOR_COL].astype(str)
            + f"<br /><br /><b>{agg_type_str}: </b>"
            + round(pdf[C.Y_AXIS_COL], 2).astype(str)
        )
    return pdf


def get_retention_tooltip(pdf: pd.DataFrame, metric: M.RetentionMetric) -> pd.DataFrame:
    agg_type_str = L.agg_type_label(metric._agg_type, metric._agg_param)
    value_suffix = "%"
    pdf[C.TOOLTIP_COL] = (
        "<b>Retention period: </b>"
        + (
            pdf[C.X_AXIS_COL]
            .apply(lambda val: C.retention_period_label(val, metric))
            .astype(str)
        )
        + "<br /><b>Cohort: </b>"
        + pdf[C.COLOR_COL].astype(str)
        + f"<br /><br /><b>{agg_type_str}: </b>"
        + round(pdf[C.Y_AXIS_COL], 2).astype(str)
        + f" {value_suffix}"
        + "<br /><b>Initial user count: </b>"
        + pdf[GA.USER_COUNT_COL + "_1"].astype(str)
        + "<br /><b>Retaining user count: </b>"
        + pdf[GA.USER_COUNT_COL + "_2"].astype(str)
    )
    return pdf



from __future__ import annotations


import mitzu.model as M
from mitzu.adapters.sqlalchemy_adapter import SQLAlchemyAdapter, FieldReference
from typing import Any, List
from snowflake.sqlalchemy.custom_types import TIMESTAMP_NTZ, TIMESTAMP_TZ
import sqlalchemy as SA
import sqlalchemy.sql.expression as EXP
import sqlalchemy.sql.sqltypes as SA_T
import mitzu.adapters.generic_adapter as GA
import pandas as pd
import ast


class SnowflakeAdapter(SQLAlchemyAdapter):
    def _get_connection_url(self, con: M.Connection):
        url = super()._get_connection_url(con)
        extra_args = {}

        if "warehouse" in con.extra_configs.keys():
            extra_args["warehouse"] = con.extra_configs["warehouse"]

        if "role" in con.extra_configs.keys():
            extra_args["role"] = con.extra_configs["role"]

        if len(extra_args) > 0:
            url += "?" + "&".join(
                f"{key}={value}" for (key, value) in extra_args.items()
            )

        return url

    def map_type(self, sa_type: Any) -> M.DataType:
        if type(sa_type) in [TIMESTAMP_NTZ, TIMESTAMP_TZ]:
            return M.DataType.DATETIME

        return super().map_type(sa_type)

    def _get_datetime_interval(
        self, field_ref: FieldReference, timewindow: M.TimeWindow
    ) -> Any:
        return SA.func.dateadd(
            timewindow.period.name.lower(),
            timewindow.value,
            field_ref,
        )

    def _get_dynamic_datetime_interval(
        self,
        field_ref: FieldReference,
        value_field_ref: FieldReference,
        time_group: M.TimeGroup,
    ) -> Any:
        return SA.func.dateadd(time_group.name.lower(), value_field_ref, field_ref)

    def _get_column_values_df(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ) -> pd.DataFrame:
        df = super()._get_column_values_df(event_data_table, fields, event_specific)
        for field in df.columns:
            df[field] = df[field].apply(
                lambda val: ast.literal_eval(val) if val is not None else None
            )
        return df

    def _get_conv_aggregation(
        self, metric: M.Metric, cte: EXP.CTE, first_cte: EXP.CTE
    ) -> Any:
        if metric._agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
            if metric._agg_param is None or 0 < metric._agg_param > 100:
                raise ValueError(
                    "Conversion percentile parameter must be between 0 and 100"
                )
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            return SA.cast(
                SA.func.approx_percentile(
                    SA.func.datediff("second", t1, t2), metric._agg_param / 100.0
                ),
                SA_T.FLOAT,
            )
        if metric._agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            return SA.cast(SA.func.avg(SA.func.datediff("second", t1, t2)), SA_T.FLOAT)
        else:
            return super()._get_conv_aggregation(metric, cte, first_cte)



from __future__ import annotations

from typing import Optional, cast

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import flask
from dash import CeleryManager, Dash, DiskcacheManager, dcc, html, page_container
from dash.long_callback.managers import BaseLongCallbackManager

import mitzu.webapp.configs as configs
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.offcanvas as OC
from mitzu.helper import LOGGER
from mitzu.webapp.helper import MITZU_LOCATION

MAIN = "main"

MDB_CSS = "https://cdnjs.cloudflare.com/ajax/libs/mdb-ui-kit/6.0.1/mdb.min.css"
DCC_DBC_CSS = (
    "https://cdn.jsdelivr.net/gh/AnnMarieW/dash-bootstrap-templates/dbc.min.css"
)


def create_webapp_layout(dependencies: DEPS.Dependencies) -> bc.Component:
    LOGGER.debug("Initializing WebApp")
    offcanvas = OC.create_offcanvas(dependencies)
    location = dcc.Location(id=MITZU_LOCATION, refresh=False)
    return html.Div(
        children=[location, offcanvas, page_container],
        className=MAIN,
        id=MAIN,
    )


def get_callback_manager(dependencies: DEPS.Dependencies) -> BaseLongCallbackManager:
    if configs.QUEUE_REDIS_HOST is not None:
        from celery import Celery

        celery_app = Celery(
            __name__, broker=configs.QUEUE_REDIS_HOST, backend=configs.QUEUE_REDIS_HOST
        )
        return CeleryManager(
            celery_app,
            cache_by=[lambda: configs.LAUNCH_UID],
            expire=configs.CACHE_EXPIRATION,
        )
    else:
        import mitzu.webapp.cache as C

        return DiskcacheManager(
            cast(C.DiskMitzuCache, dependencies.queue).get_disk_cache(),
            cache_by=[lambda: configs.LAUNCH_UID],
            expire=configs.CACHE_EXPIRATION,
        )


def create_dash_app(dependencies: Optional[DEPS.Dependencies] = None) -> Dash:
    server = flask.Flask(__name__)
    if dependencies is None:
        dependencies = DEPS.Dependencies.from_configs(server)

    with server.app_context():
        flask.current_app.config[DEPS.CONFIG_KEY] = dependencies

    app = Dash(
        __name__,
        compress=configs.DASH_COMPRESS_RESPONSES,
        server=server,
        external_stylesheets=[
            MDB_CSS,
            dbc.icons.BOOTSTRAP,
            "/assets/explore_page.css",
        ],
        assets_folder=configs.DASH_ASSETS_FOLDER,
        assets_url_path=configs.DASH_ASSETS_URL_PATH,
        serve_locally=configs.DASH_SERVE_LOCALLY,
        title=configs.DASH_TITLE,
        update_title=None,
        suppress_callback_exceptions=True,
        use_pages=True,
        background_callback_manager=get_callback_manager(dependencies),
        external_scripts=[
            "https://cdnjs.cloudflare.com/ajax/libs/dragula/3.7.2/dragula.min.js"
        ],
    )
    app._favicon = configs.DASH_FAVICON_PATH
    app.layout = create_webapp_layout(dependencies)

    @server.route(configs.HEALTH_CHECK_PATH)
    def healthcheck():
        return flask.Response("ok", status=200)

    return app


if __name__ == "__main__":
    app = create_dash_app()

    app.run(debug=True)



import traceback
from typing import List, cast

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import flask
from dash import html, register_page

import mitzu.helper as H
import mitzu.model as M
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted_layout

CONNECTIONS_CONTAINER = "connections_container"
CONNECTION_TITLE = "connection_title"


@restricted_layout
def layout() -> bc.Component:
    depenednecies: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    connection_ids = depenednecies.storage.list_connections()

    connections: List[M.Connection] = []
    for con_id in connection_ids:
        con = depenednecies.storage.get_connection(con_id)
        connections.append(con)

    return html.Div(
        [
            NB.create_mitzu_navbar("explore-navbar", []),
            dbc.Container(
                children=[
                    html.H4("Choose from connections:", className="card-title"),
                    html.Hr(),
                    create_connections_container(connections),
                    html.Hr(),
                    dbc.Row(
                        dbc.Col(
                            dbc.Button(
                                [
                                    html.B(className="bi bi-plus-circle"),
                                    " New Warehouse Connection",
                                ],
                                color="primary",
                                href=P.CONNECTIONS_CREATE_PATH,
                            ),
                            lg=3,
                            sm=12,
                        )
                    ),
                ]
            ),
        ]
    )


def create_connections_container(connections: List[M.Connection]):
    children = []

    for con in connections:
        try:
            comp = create_connection_selector(con)
            children.append(comp)
        except Exception as exc:
            traceback.print_exception(type(exc), exc, exc.__traceback__)

    if len(children) == 0:
        return html.H4(
            "You don't have any connections yet...", className="card-title text-center"
        )

    return dbc.Row(children=children, id=CONNECTIONS_CONTAINER)


def create_connection_selector(connection: M.Connection) -> bc.Component:
    details: List[str] = []
    if connection.host is not None:
        details.append(f"Host: {connection.host}")
    if connection.port is not None:
        details.append(f"Port: {connection.port}")
    if connection.catalog is not None:
        details.append(f"Catalog: {connection.catalog}")
    if connection.schema is not None:
        details.append(f"Schema: {connection.schema}")

    project_jumbotron = dbc.Col(
        dbc.Card(
            dbc.CardBody(
                [
                    html.H4(
                        H.value_to_label(connection.connection_name),
                        className="card-title",
                        id=CONNECTION_TITLE,
                    ),
                    html.Hr(),
                    html.Img(
                        src=f"/assets/warehouse/{str(connection.connection_type.name).lower()}.png",
                        height=40,
                    ),
                    *[html.Div(d) for d in details],
                    html.Hr(),
                    html.Div(
                        [
                            dbc.Button(
                                "Manage",
                                color="primary",
                                href=P.create_path(
                                    P.CONNECTIONS_MANAGE_PATH,
                                    connection_id=connection.id,
                                ),
                            ),
                        ],
                    ),
                ]
            ),
            class_name="mb-3",
        ),
        lg=3,
        sm=12,
    )
    return project_jumbotron


register_page(
    __name__,
    path=P.CONNECTIONS_PATH,
    title="Mitzu - Connections",
)



from __future__ import annotations

from typing import Any

import mitzu.adapters.generic_adapter as GA
import mitzu.model as M
import pandas as pd
from mitzu.adapters.sqlalchemy_adapter import FieldReference, SQLAlchemyAdapter

import sqlalchemy as SA
import sqlalchemy.sql.expression as EXP

NULL_VALUE_KEY = "##NULL##"


class MySQLAdapter(SQLAlchemyAdapter):
    def __init__(self, project: M.Project):
        super().__init__(project)

    def _get_distinct_array_agg_func(self, field_ref: FieldReference) -> Any:
        return SA.func.cast(
            SA.func.json_keys(
                SA.func.json_objectagg(SA.func.coalesce(field_ref, NULL_VALUE_KEY), "")
            ),
            SA.JSON,
        )

    def _get_datetime_interval(
        self, field_ref: FieldReference, timewindow: M.TimeWindow
    ) -> Any:
        return field_ref + SA.text(f"interval {timewindow.value} {timewindow.period}")

    def _get_dynamic_datetime_interval(
        self,
        field_ref: FieldReference,
        value_field_ref: FieldReference,
        time_group: M.TimeGroup,
    ) -> Any:
        return field_ref + SA.text(f"interval {value_field_ref} {time_group}")

    def _get_date_trunc(self, time_group: M.TimeGroup, field_ref: FieldReference):
        if time_group == M.TimeGroup.WEEK:
            return SA.func.date_add(
                SA.func.date(field_ref),
                EXP.text(f"interval -{SA.func.weekday(field_ref)} day"),
            )

        elif time_group == M.TimeGroup.SECOND:
            fmt = "%Y-%m-%dT%H:%i:%S"
        elif time_group == M.TimeGroup.MINUTE:
            fmt = "%Y-%m-%dT%H:%i:00"
        elif time_group == M.TimeGroup.HOUR:
            fmt = "%Y-%m-%dT%H:00:00"
        elif time_group == M.TimeGroup.DAY:
            fmt = "%Y-%m-%d"
        elif time_group == M.TimeGroup.MONTH:
            fmt = "%Y-%m-01"
        elif time_group == M.TimeGroup.QUARTER:
            raise NotImplementedError(
                "Timegroup Quarter is not supported for MySQL Adapter"
            )
        elif time_group == M.TimeGroup.YEAR:
            fmt = "%Y-01-01"

        return SA.func.timestamp(SA.func.date_format(field_ref, fmt))

    def _get_conv_aggregation(
        self, metric: M.Metric, cte: EXP.CTE, first_cte: EXP.CTE
    ) -> Any:
        if metric._agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
            raise NotImplementedError(
                "Percentile calculation is not supported by MySQL Connections."
            )
        if metric._agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            diff = SA.func.unix_timestamp(t2) - SA.func.unix_timestamp(t1)
            return SA.func.avg(diff)
        else:
            return super()._get_conv_aggregation(metric, cte, first_cte)

    def get_conversion_df(self, metric: M.ConversionMetric) -> pd.DataFrame:
        df = super().get_conversion_df(metric)
        for index in range(1, len(metric._conversion._segments) + 1):
            df[f"{GA.AGG_VALUE_COL}_{index}"] = df[
                f"{GA.AGG_VALUE_COL}_{index}"
            ].astype(float)
        return df

    def get_segmentation_df(self, metric: M.SegmentationMetric) -> pd.DataFrame:
        df = super().get_segmentation_df(metric)
        df[GA.AGG_VALUE_COL] = df[GA.AGG_VALUE_COL].astype(float)
        return df



from typing import Optional, cast

import dash.development.base_component as bc
import flask
from dash import html, register_page

import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.model as WM
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.dashboards.manage_dashboards_component as MDC
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted_layout


@restricted_layout
def no_connection_layout():
    return layout(None)


@restricted_layout
def layout(dashboard_id: Optional[str] = None, **query_params) -> bc.Component:
    dashboard: Optional[WM.Dashboard] = None
    if dashboard_id is not None:
        depenednecies: DEPS.Dependencies = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        )
        dashboard = depenednecies.storage.get_dashboard(dashboard_id)

    return html.Div(
        [
            NB.create_mitzu_navbar("create-dashboard-navbar", []),
            MDC.create_manage_dashboard_container(dashboard),
        ]
    )


register_page(
    __name__ + "_create",
    path=P.DASHBOARDS_CREATE_PATH,
    title="Mitzu - Create dashboard",
    layout=no_connection_layout,
)

register_page(
    __name__,
    path_template=P.DASHBOARDS_EDIT_PATH,
    title="Mitzu - Edit dashboard",
    layout=layout,
)



import dash_bootstrap_components as dbc
from dash import ALL, Input, Output, State, callback, ctx, html, no_update
import dash.development.base_component as bc
from typing import Callable, Dict, List, Optional, Tuple, cast, Set
import mitzu.model as M
from mitzu.webapp.pages.projects.helper import (
    PROP_CONNECTION,
    PROJECT_INDEX_TYPE,
    EDT_TBL_BODY,
    MISSING_FIELD,
    create_empty_edt,
    get_value_from_row,
)
from mitzu.webapp.auth.decorator import restricted
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.helper as H
import flask
import traceback
import dash_mantine_components as dmc


TBL_ID = "add_tables_id"
TBL_SEARCH_INPUT = "table_search_box"
TBL_PROGRESS_INFO = "table_progress_info"

TBL_CHECK_BOX_TYPE = "table_checkbox"
TBL_HEADER_CHECK_BOX_TYPE = "table_header_checkbox"

ADD_TABLES_BUTTON = "add_tables_button"
REMOVE_TABLES_BUTTON = "remove_tables_button"
CONFIGURE_TABLES_BUTTON = "configure_tables_button"

ADD_TABLES_MODAL = "add_tables_modal"
ADD_TABLES_MODAL_CLOSE = "add_tables_modal_close"
ADD_TABLES_MODAL_CONFIRM = "add_tables_modal_confirm"

CONF_TABLES_MODAL = "conf_tables_modal"
CONF_TABLES_MODAL_CLOSE = "conf_tables_modal_close"
CONF_TABLES_MODAL_CONFIRM = "conf_tables_modal_confirm"

CHOOSE_SCHEMA_INFO = "choose_schema_info"
CHOOSE_SCHEMA_DD = "choose_schema_dd"
CHOOSE_TABLES_CHECKLIST = "choose_tables_checklist"
CHOOSE_TABLES_INFO = "choose_tables_info"
SELECT_ALL_TABLES_BUTTON = "select_all_tables"

CONF_PROP_PROGRES_INFO = "conf_progree_info"
CONF_PROP_EVENT_TIME = "event_time_column"
CONF_PROP_USER_ID = "user_id_column"
CONF_PROP_EVENT_NAME_COLUMN = "event_name_column"
CONF_PROP_DATE_PARTITION = "date_partition_column"
CONF_PROP_IGNORE_COLUMN = "ignore_columns"

EDT_INDEX_TYPE = "edt_property_type"
EDT_VALIDATE_BUTTON = "edt_validate_button"
UPDATE_INTERVAL = 100


def create_table_row(edt: M.EventDataTable) -> html.Tr:
    return html.Tr(
        children=[
            html.Td(
                dbc.Checkbox(
                    id={
                        "type": TBL_CHECK_BOX_TYPE,
                        "index": edt.table_name,
                    },
                    value=False,
                ),
                className=H.TBL_CLS,
            ),
            html.Td(children=edt.get_full_name(), className=H.TBL_CLS),
            html.Td(
                children=edt.user_id_field._get_name(),
                className=H.TBL_CLS
                if edt.user_id_field != MISSING_FIELD
                else H.TBL_CLS_WARNING,
            ),
            html.Td(
                children=edt.event_time_field._get_name(),
                className=H.TBL_CLS
                if edt.event_time_field != MISSING_FIELD
                else H.TBL_CLS_WARNING,
            ),
            html.Td(
                children=(
                    edt.event_name_field._get_name() if edt.event_name_field else None
                ),
                className=H.TBL_CLS,
            ),
            html.Td(
                children=edt.date_partition_field._get_name()
                if edt.date_partition_field
                else None,
                className=H.TBL_CLS,
            ),
            html.Td(
                ",".join([f._get_name() for f in edt.ignored_fields]),
                className=H.TBL_CLS,
            ),
        ]
    )


def create_table_component(project: Optional[M.Project]):
    table_header = [
        html.Thead(
            html.Tr(
                children=[
                    html.Th(
                        dbc.Checkbox(id=TBL_HEADER_CHECK_BOX_TYPE),
                        className=H.TBL_HEADER_CLS,
                    ),
                    html.Th("Table*", className=H.TBL_HEADER_CLS),
                    html.Th("User id column*", className=H.TBL_HEADER_CLS),
                    html.Th("Event time column*", className=H.TBL_HEADER_CLS),
                    html.Th("Event name column", className=H.TBL_HEADER_CLS),
                    html.Th("Date partition column", className=H.TBL_HEADER_CLS),
                    html.Th("Ignore columns", className=H.TBL_HEADER_CLS),
                ],
            )
        )
    ]
    rows = []
    if project is not None:
        for edt in project.event_data_tables:
            rows.append(create_table_row(edt))

    table_body = [html.Tbody(children=rows, id=EDT_TBL_BODY)]

    return dbc.Table(
        children=table_header + table_body,
        hover=False,
        responsive=True,
        striped=True,
        size="sm",
        id=TBL_ID,
    )


def create_add_tables_modal() -> dbc.Modal:
    return dbc.Modal(
        [
            dbc.ModalHeader(dbc.ModalTitle("Add tables"), close_button=False),
            dbc.ModalBody(
                [
                    html.Div(
                        [
                            dbc.Label(
                                "Select schema",
                                class_name="lead w-50",
                                id=CHOOSE_SCHEMA_INFO,
                            ),
                            dmc.Select(
                                id=CHOOSE_SCHEMA_DD,
                                className="w-50 d-inline-block",
                                placeholder="Loading...",
                                data=[],
                                size="xs",
                                searchable=True,
                            ),
                        ]
                    ),
                    html.Hr(),
                    dbc.Label(
                        children=[],
                        id=CHOOSE_TABLES_INFO,
                        class_name="mb-3 d-block lead",
                    ),
                    dbc.Checklist(
                        id=CHOOSE_TABLES_CHECKLIST,
                        options=[],
                        value=[],
                        class_name="ms-3 mb-3",
                        style={"min-height": "100px"},
                        label_class_name="small",
                    ),
                    dbc.Button(
                        children=[
                            html.B(className="bi bi-check2-all me-1"),
                            "Select all",
                        ],
                        id=SELECT_ALL_TABLES_BUTTON,
                        size="sm",
                        color="secondary",
                        class_name="mb-3 d-inline-block me-1",
                        disabled=True,
                    ),
                ],
            ),
            dbc.ModalFooter(
                [
                    dbc.Button(
                        [html.B(className="bi bi-x me-1"), "Close"],
                        id=ADD_TABLES_MODAL_CLOSE,
                        size="sm",
                        color="secondary",
                        class_name="me-1",
                    ),
                    dbc.Button(
                        [html.B(className="bi bi-plus-circle me-1"), "Add tables"],
                        id=ADD_TABLES_MODAL_CONFIRM,
                        size="sm",
                        color="success",
                    ),
                ]
            ),
        ],
        id=ADD_TABLES_MODAL,
        scrollable=True,
        is_open=False,
        backdrop="static",
    )


def create_configure_tables_modal() -> dbc.Modal:
    return dbc.Modal(
        [
            dbc.ModalHeader(dbc.ModalTitle("Configure tables"), close_button=False),
            dbc.ModalBody(
                dbc.Form(
                    [
                        html.Div(
                            [
                                html.Div(
                                    "Loading table columns",
                                    className="lead d-block mb-3 d-inline-block",
                                    id=CONF_PROP_PROGRES_INFO,
                                ),
                            ]
                        ),
                        H.create_form_property_input(
                            property=CONF_PROP_USER_ID,
                            index_type=EDT_INDEX_TYPE,
                            component_type=dmc.MultiSelect,
                            data=[],
                            value=None,
                            required=True,
                            icon_cls="bi bi-person-circle",
                            label_lg=4,
                            label_sm=4,
                            input_lg=8,
                            input_sm=8,
                            size="xs",
                            searchable=True,
                            placeholder="Loading...",
                            dropdownPosition="bottom",
                        ),
                        H.create_form_property_input(
                            property=CONF_PROP_EVENT_TIME,
                            index_type=EDT_INDEX_TYPE,
                            component_type=dmc.MultiSelect,
                            data=[],
                            value=None,
                            required=True,
                            icon_cls="bi bi-clock",
                            label_lg=4,
                            label_sm=4,
                            input_lg=8,
                            input_sm=8,
                            size="xs",
                            searchable=True,
                            placeholder="Loading...",
                            dropdownPosition="bottom",
                        ),
                        H.create_form_property_input(
                            property=CONF_PROP_EVENT_NAME_COLUMN,
                            index_type=EDT_INDEX_TYPE,
                            component_type=dmc.MultiSelect,
                            data=[],
                            value=None,
                            icon_cls="bi bi-play-btn",
                            label_lg=4,
                            label_sm=4,
                            input_lg=8,
                            input_sm=8,
                            size="xs",
                            searchable=True,
                            clearable=True,
                            placeholder="Loading...",
                            dropdownPosition="bottom",
                        ),
                        H.create_form_property_input(
                            property=CONF_PROP_DATE_PARTITION,
                            index_type=EDT_INDEX_TYPE,
                            component_type=dmc.MultiSelect,
                            data=[],
                            value=None,
                            icon_cls="bi bi-calendar2-check",
                            label_lg=4,
                            label_sm=4,
                            input_lg=8,
                            input_sm=8,
                            size="xs",
                            searchable=True,
                            clearable=True,
                            placeholder="Loading...",
                            dropdownPosition="bottom",
                        ),
                        H.create_form_property_input(
                            property=CONF_PROP_IGNORE_COLUMN,
                            index_type=EDT_INDEX_TYPE,
                            component_type=dmc.MultiSelect,
                            data=[],
                            value=None,
                            icon_cls="bi bi-file-x",
                            label_lg=4,
                            label_sm=4,
                            input_lg=8,
                            input_sm=8,
                            size="xs",
                            clearable=True,
                            searchable=True,
                            placeholder="Loading...",
                            dropdownPosition="bottom",
                        ),
                        html.Div("", style={"min-height": "200px"}),
                    ],
                ),
                class_name="my-3",
            ),
            dbc.ModalFooter(
                [
                    dbc.Button(
                        [html.B(className="bi bi-x me-1"), "Close"],
                        id=CONF_TABLES_MODAL_CLOSE,
                        size="sm",
                        color="secondary",
                        class_name="me-1",
                    ),
                    dbc.Button(
                        [html.B(className="bi bi-check-circle me-1"), "Configure"],
                        id=CONF_TABLES_MODAL_CONFIRM,
                        size="sm",
                        color="success",
                    ),
                ]
            ),
        ],
        id=CONF_TABLES_MODAL,
        is_open=False,
        scrollable=True,
        backdrop="static",
    )


def create_event_tables(project: Optional[M.Project]) -> bc.Component:
    table = create_table_component(project)
    add_tables_modal = create_add_tables_modal()
    configure_tables_modal = create_configure_tables_modal()
    return html.Div(
        [
            html.P(
                children=[
                    html.I(className="bi bi-info-circle me-1"),
                    """Event tables are tables in the data warehouse that contain user events. 
                The user id and event time columns are mandatory for the event tables. 
                The event name column is optional and reserved for 
                data warehouse tables that contain multiple event types.""",
                ],
                className="mb-3 lead",
            ),
            dbc.Row(
                [
                    dbc.Col(
                        dbc.Button(
                            [html.B(className="bi bi-plus-circle me-1"), "Add tables"],
                            color="primary",
                            id=ADD_TABLES_BUTTON,
                        ),
                        width="auto",
                        class_name="mb-3",
                    ),
                    dbc.Col(
                        dbc.Button(
                            [html.B(className="bi bi-gear me-1"), "Configure"],
                            color="secondary",
                            disabled=True,
                            id=CONFIGURE_TABLES_BUTTON,
                        ),
                        class_name="mb-3",
                        width="auto",
                    ),
                    dbc.Col(
                        dbc.Button(
                            [html.B(className="bi bi-x-circle me-1"), "Remove"],
                            color="danger",
                            disabled=True,
                            id=REMOVE_TABLES_BUTTON,
                        ),
                        class_name="mb-3",
                        width="auto",
                    ),
                    dbc.Col(
                        dbc.Input(id=TBL_SEARCH_INPUT, placeholder="Search tables"),
                        class_name="ms-auto mb-3",
                        width="3",
                    ),
                ],
                class_name="me-auto",
            ),
            table,
            html.Hr(),
            dbc.Button(
                [
                    html.B(className="bi bi-check-circle me-1"),
                    "Validate",
                ],
                id=EDT_VALIDATE_BUTTON,
                color="primary",
                className="me-3 d-inline-block shadow-sm mb-3",
            ),
            html.Div(
                children=[],
                id=TBL_PROGRESS_INFO,
                className="lead d-inline-block mb-3",
            ),
            add_tables_modal,
            configure_tables_modal,
        ],
        className="overflow-auto mh-100 mt-3",
    )


def get_checkbox_value_from_row(tr: html.Tr) -> bool:
    return (
        tr.get("props")
        .get("children")[0]
        .get("props")
        .get("children")
        .get("props")["value"]
    )


def _get_unioned_table_fields(
    set_progress,
    adapter,
    all_selected_tables_count: int,
    tbl_body_children: List[bc.Component],
) -> Tuple[Dict[str, int], List[Set]]:
    """If the user selects tables with the checkboxes to configure,
    this method returns all possible fields unioned for those tables.
    Also it returns the values that were already chosen based on the table cells (user_id, event_time, etc.).
    However it cross checks them with the state of the fields from the DWH.
    """
    count = 0
    unioned_table_fields: Dict[
        str, int
    ] = {}  # for counting occurances of field in all tables
    collected_prop_values: List[Set[str]] = [set(), set(), set(), set(), set()]
    for tr in tbl_body_children:
        check_box = get_checkbox_value_from_row(tr)
        if check_box:
            table_name_children = get_value_from_row(tr, 1)
            fields_from_table = [get_value_from_row(tr, i) for i in range(2, 7)]
            if type(table_name_children) == str:
                table_parts = table_name_children.split(".")
                schema = table_parts[0]
                table_name = table_parts[-1]
                fields = adapter.list_all_table_columns(schema, table_name)
                for field in fields:
                    for f in field.get_all_subfields():
                        tbl_field_name = f._get_name()
                        val = unioned_table_fields.get(tbl_field_name)
                        unioned_table_fields[tbl_field_name] = (
                            1 if val is None else val + 1
                        )
                        for i, field in enumerate(fields_from_table):
                            if field:
                                if i < 5 and tbl_field_name == field:
                                    collected_prop_values[i].add(tbl_field_name)
                                elif tbl_field_name in field.split(","):
                                    # Ignore fields is a list of fields separated by comma
                                    collected_prop_values[i].add(tbl_field_name)

            count += 1
            set_progress(
                f"Loading table columns {count*100/all_selected_tables_count:.0f}%"
            )
    return unioned_table_fields, collected_prop_values


def find_first_field(
    all_fields: Dict[str, M.Field], field_names: List[str]
) -> Optional[M.Field]:
    for fn in field_names:
        field = all_fields.get(fn)
        if field:
            return field
    return None


@callback(
    Output(CONFIGURE_TABLES_BUTTON, "disabled"),
    Output(REMOVE_TABLES_BUTTON, "disabled"),
    Output({"type": TBL_CHECK_BOX_TYPE, "index": ALL}, "value"),
    Input(TBL_HEADER_CHECK_BOX_TYPE, "value"),
    Input({"type": TBL_CHECK_BOX_TYPE, "index": ALL}, "value"),
    State(EDT_TBL_BODY, "children"),
    prevent_initial_call=True,
)
@restricted
def manage_table_checkboxes(
    header_checkbox: bool, tr_checkboxes: List, tbl_rows: List
) -> Tuple[bool, bool, List[bool]]:
    if ctx.triggered_id == TBL_HEADER_CHECK_BOX_TYPE:
        checkboxes: List[bool] = []
        for tr in tbl_rows:
            style = tr["props"].get("style", {}).get("display", "table-row")
            if style == "table-row":
                checkboxes.append(header_checkbox)
            else:
                checkboxes.append(get_checkbox_value_from_row(tr))
        disabled = not any(checkboxes)
        return (
            disabled,
            disabled,
            checkboxes,
        )
    else:
        tr_checkboxes = [get_checkbox_value_from_row(tr) for tr in tbl_rows]
        disabled = not any(tr_checkboxes)
        return (disabled, disabled, tr_checkboxes)


@callback(
    Output(ADD_TABLES_MODAL, "is_open"),
    Input(ADD_TABLES_BUTTON, "n_clicks"),
    Input(ADD_TABLES_MODAL_CLOSE, "n_clicks"),
    Input(ADD_TABLES_MODAL_CONFIRM, "n_clicks"),
)
@restricted
def manage_add_table_modal_open(add_tables: int, close: int, confirm: int) -> bool:
    return ctx.triggered_id == ADD_TABLES_BUTTON


@callback(
    Output(CONF_TABLES_MODAL, "is_open"),
    Input(CONF_TABLES_MODAL_CLOSE, "n_clicks"),
    Input(CONF_TABLES_MODAL_CONFIRM, "n_clicks"),
    Input(CONFIGURE_TABLES_BUTTON, "n_clicks"),
)
@restricted
def manage_configure_table_modal_open(close: int, confirm: int, configure: int) -> bool:
    return ctx.triggered_id == CONFIGURE_TABLES_BUTTON


@callback(
    Output(CHOOSE_SCHEMA_DD, "data"),
    Output(CHOOSE_SCHEMA_DD, "value"),
    Output(CHOOSE_SCHEMA_DD, "placeholder"),
    Input(ADD_TABLES_BUTTON, "n_clicks"),
    Input(ADD_TABLES_MODAL_CLOSE, "n_clicks"),
    Input(ADD_TABLES_MODAL_CONFIRM, "n_clicks"),
    State({"type": PROJECT_INDEX_TYPE, "index": PROP_CONNECTION}, "value"),
    background=True,
    running=[
        (
            Output(CHOOSE_SCHEMA_DD, "placeholder"),
            "Loading schemas...",
            "Select schema",
        ),
        (
            Output(CHOOSE_SCHEMA_INFO, "children"),
            [
                dbc.Spinner(
                    spinner_style={"width": "1rem", "height": "1rem"},
                    spinner_class_name="me-1",
                ),
                "Loading schemas...",
            ],
            "Select schema:",
        ),
    ],
    interval=UPDATE_INTERVAL,
    prevent_initial_call=True,
)
@restricted
def manage_choose_schema_dropdown(
    add_tables: int, close: int, confirm: int, connection_id: Optional[str]
) -> Tuple:
    if ctx.triggered_id in [ADD_TABLES_MODAL_CLOSE, ADD_TABLES_MODAL_CONFIRM]:
        return ([], None, "Select schema")
    if connection_id is not None:
        try:
            dependencies: DEPS.Dependencies = cast(
                DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
            )
            connection = dependencies.storage.get_connection(connection_id)
            dummy_project = M.Project(
                connection=connection,
                event_data_tables=[],
                project_name="dummy_project",
            )
            adapter = dummy_project.get_adapter()
            schemas = [{"label": s, "value": s} for s in adapter.list_schemas()]
            return (schemas, no_update, "Select schema")
        except Exception:
            traceback.print_exc()
    return ([], None, "Something went wrong...")


@callback(
    Output(CHOOSE_TABLES_CHECKLIST, "options"),
    Output(CHOOSE_TABLES_CHECKLIST, "value"),
    Output(CHOOSE_TABLES_INFO, "children"),
    Input(CHOOSE_SCHEMA_DD, "value"),
    Input(SELECT_ALL_TABLES_BUTTON, "n_clicks"),
    Input(ADD_TABLES_MODAL_CLOSE, "n_clicks"),
    Input(ADD_TABLES_MODAL_CONFIRM, "n_clicks"),
    State({"type": PROJECT_INDEX_TYPE, "index": PROP_CONNECTION}, "value"),
    State(CHOOSE_TABLES_CHECKLIST, "options"),
    State(CHOOSE_TABLES_CHECKLIST, "value"),
    background=True,
    running=[
        (
            Output(CHOOSE_TABLES_INFO, "children"),
            [
                dbc.Spinner(
                    spinner_style={"width": "1rem", "height": "1rem"},
                    spinner_class_name="me-1",
                ),
                "Loading tables in schema",
            ],
            "Choose tables to add",
        ),
        (
            Output(SELECT_ALL_TABLES_BUTTON, "disabled"),
            True,
            False,
        ),
    ],
    interval=UPDATE_INTERVAL,
    cancel=[Input(ADD_TABLES_MODAL_CLOSE, "n_clicks")],
    prevent_initial_call=True,
)
@restricted
def manage_choose_tables_checklist(
    schema: Optional[str],
    select_all_clicks: int,
    close: int,
    confirm: int,
    connection_id: Optional[str],
    options: List,
    selected_values: List,
) -> Tuple:
    if ctx.triggered_id in [ADD_TABLES_MODAL_CLOSE, ADD_TABLES_MODAL_CONFIRM]:
        return ([], [], "No schema chosen")

    if schema is None:
        return ([], [], "No schema chosen")

    if ctx.triggered_id == SELECT_ALL_TABLES_BUTTON:
        if len(selected_values) == len(options):
            return (options, [], "Choose tables to add")
        else:
            vals = [o.get("value") for o in options]
            return (options, vals, "Choose tables to add")

    if connection_id is not None:
        try:
            dependencies: DEPS.Dependencies = cast(
                DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
            )
            connection = dependencies.storage.get_connection(connection_id)
            dummy_project = M.Project(
                connection=connection,
                event_data_tables=[],
                project_name="dummy_project",
            )
            adapter = dummy_project.get_adapter()

            tables = [
                {"label": s, "value": s} for s in adapter.list_tables(schema=schema)
            ]
            return (tables, [], "Choose tables to add")
        except Exception as exc:
            traceback.print_exc()
            return ([], [], f"Something went wrong: {exc}")
    return ([], [], "Connection must be set first!")


@callback(
    Output({"type": EDT_INDEX_TYPE, "index": ALL}, "data"),
    Output({"type": EDT_INDEX_TYPE, "index": ALL}, "value"),
    Output({"type": EDT_INDEX_TYPE, "index": ALL}, "placeholder"),
    Output(CONF_PROP_PROGRES_INFO, "children"),
    Input(CONFIGURE_TABLES_BUTTON, "n_clicks"),
    Input(CONF_TABLES_MODAL_CLOSE, "n_clicks"),
    Input(CONF_TABLES_MODAL_CONFIRM, "n_clicks"),
    State({"type": PROJECT_INDEX_TYPE, "index": PROP_CONNECTION}, "value"),
    State(EDT_TBL_BODY, "children"),
    background=True,
    interval=UPDATE_INTERVAL,
    progress=Output(CONF_PROP_PROGRES_INFO, "children"),
    cancel=[Input(CONF_TABLES_MODAL_CLOSE, "n_clicks")],
    prevent_initial_call=True,
)
@restricted
def manage_configure_property_inputs(
    set_progress,
    configure: int,
    close: int,
    confirm: int,
    connection_id: Optional[str],
    tbl_body_children: List,
) -> Tuple:
    if ctx.triggered_id in [CONF_TABLES_MODAL_CLOSE, CONF_TABLES_MODAL_CONFIRM]:
        # Setting the default values as we closed the dialog
        return (
            [[] for _ in range(0, 5)],
            [None for _ in range(0, 5)],
            ["Loading..." for _ in range(0, 5)],
            "Loading table column",
        )

    if connection_id is not None:
        try:
            dependencies: DEPS.Dependencies = cast(
                DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
            )
            connection = dependencies.storage.get_connection(connection_id)
            dummy_project = M.Project(
                connection=connection,
                event_data_tables=[],
                project_name="dummy_project",
            )
            all_selected_count = 0
            for tr in tbl_body_children:
                if get_checkbox_value_from_row(tr):
                    all_selected_count += 1
            all_table_fields, collected_values_list = _get_unioned_table_fields(
                set_progress,
                adapter=dummy_project.get_adapter(),
                all_selected_tables_count=all_selected_count,
                tbl_body_children=tbl_body_children,
            )
            unioned_table_fields = list(all_table_fields.items())
            unioned_table_fields = sorted(
                unioned_table_fields,
                key=lambda v: f"{str(999-v[1]*100).zfill(3)}-{v[0]}",
            )  # Sort order based on priority
            options = [
                {
                    "label": f"{k} {f'(missing from {all_selected_count-v})' if v != all_selected_count else ''}",
                    "value": k,
                }
                for k, v in unioned_table_fields
            ]

            info_text = f"Select properties for {all_selected_count} table{'s' if all_selected_count>1 else ''}:"

            return (
                [options for _ in range(0, 5)],
                [list(vals) for vals in collected_values_list],
                ["Select column", "Select column", "Optional", "Optional", "Optional"],
                info_text,
            )
        except Exception:
            traceback.print_exc()
    return (
        [[] for _ in range(0, 5)],
        [None for _ in range(0, 5)],
        [None for _ in range(0, 5)],
        "Something went wrong. Make sure your selected tables are validated.",
    )


def handle_configure_modal_confirm(
    set_progress,
    connection_id: str,
    tbl_body_children: List[bc.Component],
    edt_properties: List[bc.Component],
) -> List[bc.Component]:
    dependencies: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    connection = dependencies.storage.get_connection(connection_id)
    dummy_project = M.Project(
        connection=connection,
        event_data_tables=[],
        project_name="dummy_project",
    )
    adapter = dummy_project.get_adapter()
    results_tbl_children = []
    count = 0
    sel_count = 0

    for tr in tbl_body_children:
        if get_checkbox_value_from_row(tr):
            sel_count += 1

    for tr in tbl_body_children:
        check_box = get_checkbox_value_from_row(tr)
        if check_box:
            count += 1
            set_progress(f"Validating {count}/{sel_count} tables")
            full_table_name = get_value_from_row(tr, 1)
            schema, table_name = tuple(full_table_name.split("."))
            fields = adapter.list_all_table_columns(
                table_name=table_name, schema=schema
            )
            field_names: Dict[str, M.Field] = {}
            for field in fields:
                for f in field.get_all_subfields():
                    field_names[f._get_name()] = f

            user_id_fields = edt_properties[0]
            evnet_time_fields = edt_properties[1]
            event_name_fields = edt_properties[2]
            date_partition_fields = edt_properties[3]
            ignored_fields = edt_properties[4]
            ignored_fields = ignored_fields if ignored_fields is not None else []

            user_id_field = find_first_field(field_names, user_id_fields)
            event_time_field = find_first_field(field_names, evnet_time_fields)
            edt = M.EventDataTable(
                table_name=table_name,
                schema=schema,
                user_id_field=(user_id_field if user_id_field else MISSING_FIELD),
                event_time_field=(
                    event_time_field if event_time_field else MISSING_FIELD
                ),
                event_name_field=find_first_field(field_names, event_name_fields),
                date_partition_field=find_first_field(
                    field_names, date_partition_fields
                ),
                ignored_fields=[
                    f for fn, f in field_names.items() if fn in ignored_fields
                ],
            )
            results_tbl_children.append(create_table_row(edt))
        else:
            results_tbl_children.append(tr)
    return results_tbl_children


def handle_validate_button_clicked(
    set_progress,
    connection_id: str,
    tbl_body_children: List[bc.Component],
) -> List[bc.Component]:
    dependencies: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    connection = dependencies.storage.get_connection(connection_id)
    dummy_project = M.Project(
        connection=connection,
        event_data_tables=[],
        project_name="dummy_project",
    )
    adapter = dummy_project.get_adapter()
    sel_count = len(tbl_body_children)
    for i, tr in enumerate(tbl_body_children):
        set_progress(f"Validating {i+1}/{sel_count} tables")
        tr_children = tr["props"]["children"]
        full_table_name = get_value_from_row(tr, 1)
        user_id_col = get_value_from_row(tr, 2)
        event_time_col = get_value_from_row(tr, 3)
        date_partition_col = get_value_from_row(tr, 4)
        ignore_fields_str = get_value_from_row(tr, 5)
        schema, table_name = tuple(full_table_name.split("."))
        ignore_fields = (
            ignore_fields_str.split(",") if ignore_fields_str is not None else []
        )

        try:
            fields = adapter.list_all_table_columns(schema, table_name)
            tr_children[1]["props"]["className"] = H.TBL_CLS

            col_names: List[str] = []
            for field in fields:
                col_names.extend([f._get_name() for f in field.get_all_subfields()])

            tr_children[2]["props"]["className"] = (
                H.TBL_CLS if user_id_col in col_names else H.TBL_CLS_WARNING
            )
            tr_children[3]["props"]["className"] = (
                H.TBL_CLS if event_time_col in col_names else H.TBL_CLS_WARNING
            )
            tr_children[4]["props"]["className"] = (
                H.TBL_CLS if date_partition_col in col_names else H.TBL_CLS_WARNING
            )
            tr_children[5]["props"]["className"] = (
                H.TBL_CLS
                if len([f for f in ignore_fields if f not in col_names]) == 0
                else H.TBL_CLS_WARNING
            )
        except Exception:
            traceback.print_exc()
            for i in range(1, 6):
                tr_children[i]["props"]["className"] = H.TBL_CLS_WARNING

    return tbl_body_children


@callback(
    Output(EDT_TBL_BODY, "children"),
    Input(ADD_TABLES_MODAL_CONFIRM, "n_clicks"),
    Input(CONF_TABLES_MODAL_CONFIRM, "n_clicks"),
    Input(REMOVE_TABLES_BUTTON, "n_clicks"),
    Input(TBL_SEARCH_INPUT, "value"),
    Input(EDT_VALIDATE_BUTTON, "n_clicks"),
    State(CHOOSE_SCHEMA_DD, "value"),
    State(CHOOSE_TABLES_CHECKLIST, "value"),
    State(EDT_TBL_BODY, "children"),
    State({"type": EDT_INDEX_TYPE, "index": ALL}, "value"),
    State({"type": PROJECT_INDEX_TYPE, "index": PROP_CONNECTION}, "value"),
    prevent_initial_call=True,
    background=True,
    running=[
        (
            Output(TBL_ID, "className"),
            "opacity-50",
            "",
        ),
    ],
    progress=Output(TBL_PROGRESS_INFO, "children"),
    interval=UPDATE_INTERVAL,
)
@restricted
def manage_event_data_table_body(
    set_progress: Callable,
    add_tables: int,
    configure_table: int,
    remove_tables: int,
    search_value: str,
    validate_nclicks: int,
    choose_schema_dd_value: str,
    tables: List,
    tbl_body_children: List,
    edt_properties: List,
    connection_id: str,
) -> List:
    set_progress("")
    if ctx.triggered_id == REMOVE_TABLES_BUTTON:
        return [
            row for row in tbl_body_children if not get_checkbox_value_from_row(row)
        ]

    if ctx.triggered_id == EDT_VALIDATE_BUTTON:
        return handle_validate_button_clicked(
            set_progress=set_progress,
            connection_id=connection_id,
            tbl_body_children=tbl_body_children,
        )

    if ctx.triggered_id == CONF_TABLES_MODAL_CONFIRM:
        return handle_configure_modal_confirm(
            set_progress=set_progress,
            connection_id=connection_id,
            tbl_body_children=tbl_body_children,
            edt_properties=edt_properties,
        )

    if ctx.triggered_id == TBL_SEARCH_INPUT:
        for tr in tbl_body_children:
            full_table_name = get_value_from_row(tr, 1)
            hidden = search_value not in full_table_name
            tr["props"]["style"] = {"display": "none" if hidden else "table-row"}

    if ctx.triggered_id == ADD_TABLES_MODAL_CONFIRM:
        already_present: List[str] = []
        for tr in tbl_body_children:
            already_present.append(get_value_from_row(tr, 1))

        for tbl in tables:
            if f"{choose_schema_dd_value}.{tbl}" in already_present:
                continue
            edt = create_empty_edt(choose_schema_dd_value, tbl)
            tbl_body_children.append(create_table_row(edt))
    return tbl_body_children



import dash_bootstrap_components as dbc
from dash import Input, Output, callback, html
import dash.development.base_component as bc
from typing import Optional
import mitzu.model as M
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted
from mitzu.webapp.helper import create_form_property_input
from mitzu.helper import value_to_label
from mitzu.webapp.pages.projects.helper import PROP_CONNECTION, PROJECT_INDEX_TYPE
import mitzu.webapp.pages.projects.event_tables_config as ETC
import dash_mantine_components as dmc
from uuid import uuid4

CREATE_PROJECT_DOCS_LINK = "https://github.com/mitzu-io/mitzu/blob/main/DOCS.md"
PROJECT_DETAILS_CONTAINER = "project-details-container"


PROP_PROJECT_ID = "project_id"
PROP_PROJECT_NAME = "project_name"
PROP_DESCRIPTION = "description"

PROP_DISC_LOOKBACK_DAYS = "lookback_days"
PROP_DISC_SAMPLE_SIZE = "sample_size"

PROP_EXPLORE_AUTO_REFRESH = "auto_refresh"
PROP_END_DATE_CONFIG = "default_end_date_config"
PROP_CUSTOM_END_DATE_CONFIG = "custom_default_end_date"


def create_project_settings(
    project: Optional[M.Project], dependencies: DEPS.Dependencies
) -> bc.Component:
    return html.Div(
        [
            dbc.Form(
                children=[
                    dbc.Accordion(
                        children=[
                            dbc.AccordionItem(
                                children=[
                                    html.P(
                                        [
                                            html.I(className="bi bi-gear me-1"),
                                            "Setup your project:",
                                        ],
                                        className="mb-3 lead",
                                    ),
                                    create_basic_project_settings(
                                        project, dependencies
                                    ),
                                ],
                                title="Project Settings",
                            ),
                            dbc.AccordionItem(
                                create_explore_settings(project),
                                title="Explore Settings",
                            ),
                            dbc.AccordionItem(
                                create_discovery_settings(project),
                                title="Discovery Settings",
                            ),
                            dbc.AccordionItem(
                                ETC.create_event_tables(project), title="Event Tables"
                            ),
                        ],
                    ),
                ],
                id=PROJECT_DETAILS_CONTAINER,
                class_name="mt-3 ",
            ),
        ],
    )


def create_basic_project_settings(
    project: Optional[M.Project], dependencies: DEPS.Dependencies
) -> bc.Component:
    if project is not None:
        project_name = project.project_name
    else:
        project_name = None

    con_ids = dependencies.storage.list_connections()
    connections = [dependencies.storage.get_connection(c_id) for c_id in con_ids]
    con_options = [{"label": c.connection_name, "value": c.id} for c in connections]

    return html.Div(
        [
            html.Div(
                create_form_property_input(
                    property=PROP_PROJECT_ID,
                    index_type=PROJECT_INDEX_TYPE,
                    icon_cls="bi bi-info-circle",
                    value=(project.id if project is not None else str(uuid4())[-12:]),
                    disabled=True,
                ),
                className="d-none",
            ),
            create_form_property_input(
                property=PROP_PROJECT_NAME,
                index_type=PROJECT_INDEX_TYPE,
                icon_cls="bi bi-card-text",
                value=project_name,
                required=True,
                minlength=4,
                maxlength=100,
            ),
            create_form_property_input(
                property=PROP_CONNECTION,
                index_type=PROJECT_INDEX_TYPE,
                icon_cls="bi bi-link",
                component_type=dmc.Select,
                value=(project.connection.id if project is not None else None),
                data=con_options,
                size="xs",
                required=True,
            ),
            dbc.Row(
                [
                    dbc.Col("", lg=3, sm=12, class_name="m-0"),
                    dbc.Col(
                        dbc.Button(
                            [
                                html.B(className="bi bi-plus-circle me-1"),
                                "New connection",
                            ],
                            href=P.CONNECTIONS_CREATE_PATH,
                            color="primary",
                            class_name="mb-3",
                            size="sm",
                        ),
                        lg=3,
                        sm=12,
                    ),
                ]
            ),
            create_form_property_input(
                property=PROP_DESCRIPTION,
                value=project.description if project else None,
                index_type=PROJECT_INDEX_TYPE,
                component_type=dbc.Textarea,
                icon_cls="bi bi-blockquote-left",
                placeholder="Describe the project!",
                rows=4,
                maxlength=300,
            ),
        ],
    )


def create_discovery_settings(project: Optional[M.Project]) -> bc.Component:
    if project is not None:
        disc_settings = project.discovery_settings
    else:
        disc_settings = M.DiscoverySettings()
    return html.Div(
        [
            html.P(
                children=[
                    html.I(className="bi bi-info-circle me-1"),
                    """Discovery settings are needed for the project discovery step. 
                    The discovery step scrapes the data warehouse tables for the columns and it's values.""",
                ],
                className="mb-3 lead",
            ),
            create_form_property_input(
                property=PROP_DISC_LOOKBACK_DAYS,
                index_type=PROJECT_INDEX_TYPE,
                icon_cls="bi bi-clock-history",
                value=disc_settings.lookback_days,
                required=True,
                type="number",
                min=1,
            ),
            create_form_property_input(
                property=PROP_DISC_SAMPLE_SIZE,
                index_type=PROJECT_INDEX_TYPE,
                icon_cls="bi bi-filter-square",
                value=disc_settings.min_property_sample_size,
                required=True,
                type="number",
                min=1,
            ),
        ]
    )


def create_explore_settings(project: Optional[M.Project]) -> bc.Component:
    if project is not None:
        webapp_settings = project.webapp_settings
    else:
        webapp_settings = M.WebappSettings()

    return html.Div(
        [
            html.P(
                children=[
                    html.I(className="bi bi-info-circle me-1"),
                    "Explore settings change the default behaviour of the explore page.",
                ],
                className="mb-3 lead",
            ),
            create_form_property_input(
                property=PROP_EXPLORE_AUTO_REFRESH,
                index_type=PROJECT_INDEX_TYPE,
                icon_cls="bi bi-arrow-clockwise",
                value=webapp_settings.auto_refresh_enabled,
                component_type=dbc.Checkbox,
            ),
            create_form_property_input(
                property=PROP_END_DATE_CONFIG,
                index_type=PROJECT_INDEX_TYPE,
                icon_cls="bi bi-calendar2-check",
                value=webapp_settings.end_date_config.name.upper(),
                component_type=dmc.Select,
                data=[
                    {"label": value_to_label(v.name), "value": v.name.upper()}
                    for v in M.WebappEndDateConfig
                ],
                size="xs",
            ),
            create_form_property_input(
                property=PROP_CUSTOM_END_DATE_CONFIG,
                index_type=PROJECT_INDEX_TYPE,
                icon_cls="bi bi-calendar3",
                inputFormat="YYYY-MM-DD",
                value=webapp_settings.custom_end_date,
                disabled=(
                    webapp_settings.end_date_config is None
                    or webapp_settings.end_date_config
                    != M.WebappEndDateConfig.CUSTOM_DATE
                ),
                component_type=dmc.DatePicker,
                size="xs",
            ),
        ]
    )


@callback(
    Output(
        {"type": PROJECT_INDEX_TYPE, "index": PROP_CUSTOM_END_DATE_CONFIG}, "disabled"
    ),
    Input({"type": PROJECT_INDEX_TYPE, "index": PROP_END_DATE_CONFIG}, "value"),
    prevent_initial_call=True,
)
@restricted
def end_date_config_changed(config_value: str):
    return config_value.upper() != M.WebappEndDateConfig.CUSTOM_DATE.name.upper()



import dash_bootstrap_components as dbc
from dash import Input, Output, callback, ctx, html, register_page
import dash.development.base_component as bc
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
from typing import Dict, List, Tuple, Optional
from mitzu.webapp.auth.decorator import restricted, restricted_layout
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.helper as H
import flask
import mitzu.model as M
from typing import cast, Callable
import traceback
import dash_mantine_components as dmc

SELECT_PROJECT_DD = "events_select_project"
DISCOVER_INFO = "events_discovered_info"
DISCOVER_PROJECT_BUTTON = "events_discover_button"
DISCOVER_CANCEL_BUTTON = "cancel_discovery_button"
DISCOVERY_SPINNER = "discovery_spinner"

MANAGE_PROJECT_BUTTON = "discovery_manage_project_button"
DEFAULT_INTERVAL = 100

EVENTS_TBL_BODY = "events_table_body"
EVENTS_TBL_ID = "events_table_id"


def create_failed_table_row(edt: M.EventDataTable, exc: Exception) -> html.Tr:

    return html.Tr(
        [
            html.Td(edt.get_full_name(), className=H.TBL_CLS_WARNING),
            html.Td("Failed to discover table", className=H.TBL_CLS_WARNING),
            html.Td(str(exc), className=H.TBL_CLS_WARNING),
        ]
    )


def create_table_row(
    edt: M.EventDataTable, event_def: M.Reference[M.EventDef]
) -> html.Tr:
    all_fields: List[str] = []
    for field in event_def.get_value_if_exists()._fields.keys():
        all_fields.extend(sf._get_name() for sf in field.get_all_subfields())
    properties = f"{len(all_fields)} properties"

    return html.Tr(
        [
            html.Td(edt.get_full_name(), H.TBL_CLS),
            html.Td(event_def.get_value_if_exists()._event_name, className=H.TBL_CLS),
            html.Td(properties, className=H.TBL_CLS + " w-50"),
        ]
    )


def create_event_table_component(project: Optional[M.Project]) -> bc.Component:
    rows = []
    if project is not None:
        dp = project._discovered_project.get_value()
        if dp is not None:
            for edt, df in dp.definitions.items():
                for evt_df in df.values():
                    rows.append(create_table_row(edt, evt_df))

    return dbc.Table(
        children=[
            html.Thead(
                html.Tr(
                    [
                        html.Th("Source", className=H.TBL_HEADER_CLS),
                        html.Th("Event ", className=H.TBL_HEADER_CLS),
                        html.Th("Properties", className=H.TBL_HEADER_CLS),
                    ],
                )
            ),
            html.Tbody(rows, id=EVENTS_TBL_BODY),
        ],
        hover=False,
        responsive=True,
        striped=True,
        size="sm",
        id=EVENTS_TBL_ID,
    )


def no_project_layout():
    return layout(None)


@restricted_layout
def layout(project_id: Optional[str], **query_params) -> bc.Component:
    events_service = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).events_service

    projects = events_service.list_all_projects()
    selected_project: Optional[M.Project] = None
    for p in projects:
        if p.id == project_id:
            selected_project = p
            break

    options = [{"label": p.project_name, "value": p.id} for p in projects]

    return html.Div(
        [
            NB.create_mitzu_navbar("events-navbar", []),
            dbc.Container(
                [
                    html.H4("Discovered events and properties"),
                    html.Hr(),
                    dbc.Row(
                        [
                            dbc.Col("Select project:", className="lead", width="auto"),
                            dbc.Col(
                                dmc.Select(
                                    id=SELECT_PROJECT_DD,
                                    data=options,
                                    value=project_id,
                                    searchable=True,
                                    placeholder="Select project",
                                ),
                                width="2",
                            ),
                            dbc.Col(
                                dbc.Button(
                                    children=[
                                        html.B(className="bi bi-gear me-1"),
                                        "Manage project",
                                    ],
                                    id=MANAGE_PROJECT_BUTTON,
                                    color="light",
                                    disabled=project_id is None,
                                ),
                                width="auto me-auto",
                            ),
                            dbc.Col(
                                dbc.Button(
                                    [
                                        html.B(className="bi bi-x-circle me-1"),
                                        "Cancel",
                                    ],
                                    id=DISCOVER_CANCEL_BUTTON,
                                    color="light",
                                    class_name="d-inline-block mb-3",
                                ),
                                width="auto",
                                class_name="invisible",
                            ),
                            dbc.Col(
                                dbc.Button(
                                    [
                                        html.B(className="bi bi-search me-1"),
                                        "Discover project",
                                    ],
                                    id=DISCOVER_PROJECT_BUTTON,
                                    disabled=project_id is None,
                                    class_name="d-inline-block mb-3 me-3",
                                ),
                                width="auto",
                            ),
                        ],
                    ),
                    html.Hr(),
                    html.Div(
                        [
                            dbc.Spinner(
                                spinner_class_name="d-none",
                                spinner_style={"width": "1rem", "height": "1rem"},
                                id=DISCOVERY_SPINNER,
                            ),
                            html.Div(
                                children="No project selected",
                                id=DISCOVER_INFO,
                                className="mb-3 lead d-inline-block",
                            ),
                        ]
                    ),
                    create_event_table_component(selected_project),
                    html.Hr(),
                ],
            ),
        ],
    )


@callback(
    Output(MANAGE_PROJECT_BUTTON, "disabled"),
    Output(MANAGE_PROJECT_BUTTON, "href"),
    Input(SELECT_PROJECT_DD, "value"),
)
@restricted
def manage_project_disabled(project_id: str) -> Tuple[bool, str]:
    return (
        project_id is None,
        P.create_path(P.PROJECTS_MANAGE_PATH, project_id=project_id)
        if project_id is not None
        else "",
    )


@callback(
    Output(EVENTS_TBL_BODY, "children"),
    Output(DISCOVER_INFO, "children"),
    Input(DISCOVER_PROJECT_BUTTON, "n_clicks"),
    Input(SELECT_PROJECT_DD, "value"),
    background=True,
    running=[
        (Output(DISCOVER_PROJECT_BUTTON, "disabled"), True, False),
        (Output(SELECT_PROJECT_DD, "disabled"), True, False),
        (Output(DISCOVER_CANCEL_BUTTON, "class_name"), "visible", "invisible"),
        (
            Output(DISCOVERY_SPINNER, "spinner_class_name"),
            "me-1 d-inline-block",
            "d-none",
        ),
    ],
    progress=[Output(EVENTS_TBL_BODY, "children"), Output(DISCOVER_INFO, "children")],
    prevent_initial_call=True,
    interval=DEFAULT_INTERVAL,
    cancel=Input(DISCOVER_CANCEL_BUTTON, "n_clicks"),
)
@restricted
def handle_project_discovery(
    set_progress: Callable, discovery_clicks: int, project_id: str
):
    rows: List[bc.Component] = []
    events_service = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).events_service

    try:
        if ctx.triggered_id == SELECT_PROJECT_DD:
            set_progress(([], ""))
            defs = events_service.get_project_definition(project_id)
            for edt, df in defs.items():
                for evt_df in df.values():
                    rows.append(create_table_row(edt, evt_df))
        else:

            def edt_callback(
                edt: M.EventDataTable,
                defs: Dict[str, M.Reference[M.EventDef]],
                exc: Optional[Exception],
                processed_count: int,
                total_count: int,
            ):
                if exc is None:
                    for df in defs.values():
                        rows.append(create_table_row(edt, df))
                else:
                    rows.append(create_failed_table_row(edt, exc))
                set_progress(
                    (
                        rows,
                        f"Discovering tables {processed_count*100/total_count:.0f}%",
                    )
                )

            events_service.discover_project(project_id, callback=edt_callback)

        return (rows, "")

    except Exception as exc:
        traceback.print_exc()
        return ([], f"Something went wrong: {exc}")


register_page(
    __name__ + "_project",
    path=P.EVENTS_AND_PROPERTIES_PATH,
    title="Mitzu - Events And Properties",
    layout=no_project_layout,
)

register_page(
    __name__,
    path_template=P.EVENTS_AND_PROPERTIES_PROJECT_PATH,
    title="Mitzu - Events And Properties",
    layout=layout,
)



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any, Tuple, cast

import mitzu.model as M

MAX_TITLE_LENGTH = 300
MAX_SEGMENT_LENGTH = 150
DT_FORMAT = "%Y-%m-%d %H:%M:%S"
DATE_FORMAT = "%Y-%m-%d"


def format_date(metric: M.Metric, dt: datetime, adjust_rounding: int = 0) -> str:
    if metric._time_group in [M.TimeGroup.HOUR, M.TimeGroup.MINUTE, M.TimeGroup.SECOND]:
        return dt.strftime(DT_FORMAT)
    else:
        return (dt + timedelta(days=adjust_rounding)).strftime(DATE_FORMAT)


def get_timeframe_str(metric: M.Metric) -> str:
    if metric._config.start_dt is None:
        if metric._config.end_dt is None:
            return f"latest {metric._lookback_days}"
        else:
            return f"last <b>{metric._lookback_days}</b> days before <b>{format_date(metric,metric._end_dt,1)}</b>"
    else:
        return (
            f"between <b>{format_date(metric,metric._start_dt)}</b> "
            f"and <b>{format_date(metric,metric._end_dt,1)}</b>"
        )


def get_grouped_by_str(metric: M.Metric) -> str:
    if metric._group_by:
        grp = metric._group_by._field._name
        return f"grouped by <b>{grp}</b> (top {metric._max_group_count})"
    return ""


def fix_title_text(title_text: str, max_length=MAX_SEGMENT_LENGTH) -> str:
    if len(title_text) > max_length:
        return title_text[:max_length] + "..."
    else:
        return title_text


def fix_operators(op: M.Operator, right: Any) -> Tuple[M.Operator, Any]:
    if op == M.Operator.ANY_OF and type(right) == tuple and len(right) == 1:
        return M.Operator.EQ, right[0]
    elif op == M.Operator.NONE_OF and type(right) == tuple and len(right) == 1:
        return M.Operator.NEQ, right[0]
    else:
        return op, right


def get_segment_title_text(segment: M.Segment) -> str:
    if isinstance(segment, M.SimpleSegment):
        s = cast(M.SimpleSegment, segment)
        if s._operator is None:
            return f"<b>{s._left._event_name}</b>"
        else:
            left = cast(M.EventFieldDef, s._left)
            right = s._right
            operator, right = fix_operators(s._operator, right)
            if right is None:
                right = "null"

            return f"<b>{left._event_name}</b> with <b>{left._field._name}</b> {operator} <b>{right}</b>"
    elif isinstance(segment, M.ComplexSegment):
        c = cast(M.ComplexSegment, segment)
        return f"{get_segment_title_text(c._left)} {c._operator} {get_segment_title_text(c._right)}"
    else:
        raise ValueError(f"Segment of type {type(segment)} is not supported.")


def get_time_group_text(time_group: M.TimeGroup) -> str:
    if time_group == M.TimeGroup.TOTAL:
        return ""
    if time_group == M.TimeGroup.DAY:
        return "daily"
    if time_group == M.TimeGroup.MINUTE:
        return "minute by minute"

    return time_group.name.lower() + "ly"


def get_segmentation_title(metric: M.SegmentationMetric):
    if metric._config.custom_title is not None:
        return metric._config.custom_title
    segment_str = fix_title_text(get_segment_title_text(metric._segment))
    tg = get_time_group_text(metric._time_group).title()
    lines = [
        f"{tg} count of unique users",
        f"who did {segment_str}",
    ]
    if metric._group_by is not None:
        lines.append(get_grouped_by_str(metric))
    lines.append(get_timeframe_str(metric))
    return "<br />".join(lines)


def get_conversion_title(metric: M.ConversionMetric) -> str:
    if metric._config.custom_title is not None:
        return metric._config.custom_title
    events = " then did ".join(
        [
            f"{fix_title_text(get_segment_title_text(seg), 100)}"
            for seg in metric._conversion._segments
        ]
    )

    tg = get_time_group_text(metric._time_group).title()

    if metric._agg_type == M.AggType.CONVERSION:
        agg = "conversion rate"
    elif metric._agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
        if metric._agg_param == 50:
            agg = "median conversion time of users"
        else:
            agg = f"P{metric._agg_param:.0f} conversion time of users"
    elif metric._agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
        agg = "average conversion time of users"

    within_str = f"within {metric._conv_window}"
    group_by = get_grouped_by_str(metric)
    timeframe_str = get_timeframe_str(metric)

    lines = [
        f"{tg} {agg}",
        f"who did {events}",
        f"{within_str}, {group_by}",
        timeframe_str,
    ]

    return "<br />".join(lines).strip().capitalize()


def get_retention_title(metric: M.RetentionMetric):
    if metric._config.custom_title is not None:
        return metric._config.custom_title

    initial_title = get_segment_title_text(metric._initial_segment)
    retaining_title = get_segment_title_text(metric._retaining_segment)

    tg = get_time_group_text(metric._time_group).title()
    lines = [f"{tg} retention rate of users who did"]
    if initial_title == retaining_title:  # same segment
        lines += ["recurring " + initial_title]
    else:
        lines += ["recurring " + retaining_title + " after " + initial_title]
    lines += [
        f"with {metric._retention_window} periods, {get_grouped_by_str(metric)}",
        get_timeframe_str(metric),
    ]

    return "<br />".join(lines).strip().capitalize()



from __future__ import annotations

import base64
import json
import zlib
from datetime import datetime
from typing import Any, Dict, Optional, Union
import mitzu.model as M

EFD_OR_ED_TYPE = Union[M.EventFieldDef, M.EventDef]


def find_edt(project: M.Project, event_name: str) -> M.EventDataTable:
    dd = project._discovered_project.get_value()
    if dd is None:
        raise ValueError("Project not yet discovered")
    for edt, events in dd.definitions.items():
        if event_name in events.keys():
            return edt
    raise Exception(f"Couldn't find {event_name} in any of the datasources.")


def _to_dict(value: Any) -> Any:
    if value is None:
        return None
    if type(value) in (str, float, int, bool):
        return value
    if type(value) == datetime:
        return value.isoformat()
    if type(value) == list:
        return [_to_dict(v) for v in value]
    if type(value) == tuple:
        return [_to_dict(v) for v in value]

    if isinstance(value, M.TimeGroup):
        return value.name.lower()
    if isinstance(value, M.Resolution):
        return value.name.lower()
    if isinstance(value, M.TimeWindow):
        return f"{value.value} {value.period.name.lower()}"
    if isinstance(value, M.SimpleChartType):
        return value.name.lower()
    if isinstance(value, M.Field):
        return value._get_name()
    if isinstance(value, M.MetricConfig):
        res = {
            "sdt": _to_dict(value.start_dt),
            "edt": _to_dict(value.end_dt),
            "lbd": _to_dict(value.lookback_days),
            "tg": _to_dict(value.time_group),
            "mgc": _to_dict(value.max_group_count),
            "gb": _to_dict(value.group_by),
            "ct": _to_dict(value.custom_title),
            "cat": _to_dict(value.chart_type),
            "res": _to_dict(value.resolution),
        }
        if value.agg_type is not None:
            res["at"] = value.agg_type.to_agg_str(value.agg_param)
        return res
    if isinstance(value, M.EventDef):
        return {"en": value._event_name}
    if isinstance(value, M.EventFieldDef):
        return {"en": value._event_name, "f": _to_dict(value._field)}
    if isinstance(value, M.SimpleSegment):
        return {
            "l": _to_dict(value._left),
            "op": value._operator.name if value._operator is not None else None,
            "r": _to_dict(value._right) if value._right is not None else None,
        }
    if isinstance(value, M.ComplexSegment):
        return {
            "l": _to_dict(value._left),
            "bop": value._operator.name,
            "r": _to_dict(value._right),
        }
    if isinstance(value, M.Conversion):
        return {
            "segs": [_to_dict(seg) for seg in value._segments],
        }
    if isinstance(value, M.ConversionMetric):
        return {
            "conv": _to_dict(value._conversion),
            "cw": _to_dict(value._conv_window),
            "co": _to_dict(value._config),
        }
    if isinstance(value, M.RetentionMetric):
        return {
            "seg_1": _to_dict(value._initial_segment),
            "seg_2": _to_dict(value._retaining_segment),
            "rw": _to_dict(value._retention_window),
            "co": _to_dict(value._config),
        }
    if isinstance(value, M.SegmentationMetric):
        return {
            "seg": _to_dict(value._segment),
            "co": _to_dict(value._config),
        }
    raise ValueError("Unsupported value type: {}".format(type(value)))


def remove_nones(dc: Dict) -> Optional[Dict]:
    res = {}
    for k, v in dc.items():
        if type(v) == dict:
            v = remove_nones(v)
        if type(v) == list:
            v = [remove_nones(i) if type(i) == dict else i for i in v if i is not None]
        if v is not None:
            res[k] = v
    if len(res) == 0:
        return None
    return res


def to_dict(metric: M.Metric) -> Dict:
    res = _to_dict(metric)
    res = remove_nones(res)
    if res is None:
        raise Exception("Serialized metric definition is empty")
    return res


def _from_dict(
    value: Any,
    project: M.Project,
    type_hint: Any,
    path: str,
    other_hint: Any = None,
) -> Any:
    if value is None:
        return None
    if type_hint is None:
        if "conv" in value:
            return _from_dict(value, project, M.ConversionMetric, path)
        if "rw" in value:
            return _from_dict(value, project, M.RetentionMetric, path)
        if "seg" in value:
            return _from_dict(value, project, M.SegmentationMetric, path)
        if "segs" in value:
            return _from_dict(value, project, M.Conversion, path)
        if "l" in value:
            return _from_dict(value, project, M.Segment, path)
        raise ValueError(
            f"Can't deserialize metric from {value} 'conv' or 'seg' expected at {path}"
        )
    if type_hint == M.MetricConfig:
        if "at" in value:
            at, ap = _from_dict(value.get("at"), project, M.AggType, path + ".at")
        else:
            at, ap = None, None
        return M.MetricConfig(
            start_dt=_from_dict(value.get("sdt"), project, datetime, path + ".sdt"),
            end_dt=_from_dict(value.get("edt"), project, datetime, path + ".edt"),
            lookback_days=_from_dict(
                value.get("lbd"), project, M.TimeWindow, path + ".lbd"
            ),
            time_group=_from_dict(value.get("tg"), project, M.TimeGroup, path + ".tg"),
            max_group_count=_from_dict(value.get("mgc"), project, int, path + ".mgc"),
            group_by=_from_dict(
                value.get("gb"), project, M.EventFieldDef, path + ".gb"
            ),
            custom_title=_from_dict(value.get("ct"), project, str, path + ".ct"),
            resolution=_from_dict(
                value.get("res"), project, M.Resolution, path + ".res"
            ),
            chart_type=_from_dict(
                value.get("cat"), project, M.SimpleChartType, path + ".cat"
            ),
            agg_type=at,
            agg_param=ap,
        )
    if type_hint == M.ConversionMetric:
        return M.ConversionMetric(
            conversion=_from_dict(
                value.get("conv"), project, M.Conversion, path + ".conv"
            ),
            conv_window=_from_dict(
                value.get("cw"), project, M.TimeWindow, path + ".cw"
            ),
            config=_from_dict(value.get("co"), project, M.MetricConfig, path + ".co"),
        )
    if type_hint == M.Conversion:
        return M.Conversion(
            segments=_from_dict(
                value.get("segs"), project, list, path + ".segs", M.Segment
            ),
        )
    if type_hint == M.RetentionMetric:
        return M.RetentionMetric(
            initial_segment=_from_dict(
                value.get("seg_1"), project, M.Segment, path + ".seg_1"
            ),
            retaining_segment=_from_dict(
                value.get("seg_2"), project, M.Segment, path + ".seg_2"
            ),
            config=_from_dict(value.get("co"), project, M.MetricConfig, path + ".co"),
            retention_window=_from_dict(
                value.get("rw"), project, M.TimeWindow, path + ".rw"
            ),
        )
    if type_hint == M.SegmentationMetric:
        return M.SegmentationMetric(
            segment=_from_dict(value.get("seg"), project, M.Segment, path + ".seg"),
            config=_from_dict(value.get("co"), project, M.MetricConfig, path + ".co"),
        )
    if type_hint == M.Segment:
        if "bop" in value:
            return M.ComplexSegment(
                _left=_from_dict(value.get("l"), project, M.Segment, path + ".l"),
                _operator=_from_dict(
                    value.get("bop"), project, M.BinaryOperator, path + ".bop"
                ),
                _right=_from_dict(value.get("r"), project, M.Segment, path + ".r"),
            )
        elif "op" in value:
            return M.SimpleSegment(
                _left=_from_dict(value.get("l"), project, EFD_OR_ED_TYPE, path + ".l"),
                _operator=_from_dict(
                    value.get("op"), project, M.Operator, path + ".op"
                ),
                _right=_from_dict(value.get("r"), project, Any, path + ".r"),
            )
        return M.SimpleSegment(
            _left=_from_dict(value.get("l"), project, EFD_OR_ED_TYPE, path + ".l")
        )

    if type_hint == M.EventFieldDef:
        event_name = value.get("en")
        edt = find_edt(project, event_name)
        return M.EventFieldDef(
            _event_name=_from_dict(event_name, project, str, path + ".en"),
            _field=_from_dict(
                value.get("f"), project, M.Field, path + ".f", other_hint=event_name
            ),
            _event_data_table=edt,
        )
    if type_hint == M.EventDef:
        event_name = value.get("en")
        edt = find_edt(project, event_name)
        return M.EventDef(
            _event_name=_from_dict(event_name, project, str, path + ".en"),
            _fields={},
            _event_data_table=edt,
        )
    if type_hint == EFD_OR_ED_TYPE:
        if "f" in value:
            return _from_dict(value, project, M.EventFieldDef, path)
        else:
            return _from_dict(value, project, M.EventDef, path)
    if type_hint == M.Operator:
        return M.Operator[value.upper()]
    if type_hint == M.BinaryOperator:
        return M.BinaryOperator[value.upper()]
    if type_hint == M.Field:
        event_name = other_hint
        edt = find_edt(project, event_name)
        dd = project._discovered_project.get_value()
        if dd is None:
            raise Exception(f"Couldn't find {event_name} in any of the datasources.")

        event_def: M.EventDef = dd.definitions[edt][event_name].get_value_if_exists()
        fields = event_def._fields.keys()
        for f in fields:
            if f._get_name() == value:
                return f
        raise Exception(f"Couldn't find {value} among {event_name} fields.")
    if type_hint == M.TimeGroup:
        return M.TimeGroup.parse(value)
    if type_hint == M.Resolution:
        return M.Resolution.parse(value)
    if type_hint == M.AggType:
        return M.AggType.parse_agg_str(value)
    if type_hint == M.TimeWindow:
        return M.TimeWindow.parse(value)
    if type_hint == M.SimpleChartType:
        return M.SimpleChartType.parse(value)
    if type_hint == datetime:
        return datetime.fromisoformat(value)
    if type_hint == list:
        return [
            _from_dict(val, project, other_hint, path + f"[{i}]")
            for i, val in enumerate(value)
        ]
    if type_hint in [int, str, Any]:
        return value

    raise ValueError(f"Can't process {value} at {path} with type_hint {type_hint}")


def from_dict(value: Dict, project: M.Project) -> M.Metric:
    return _from_dict(value, project, None, "")


def dict_to_compressed_string(metric_dict: Dict) -> str:
    metric_json = json.dumps(metric_dict)
    zipped = zlib.compress(metric_json.encode("utf-8"), 9)
    return base64.b64encode(zipped).decode("utf-8")


def to_compressed_string(metric: M.Metric) -> str:
    return dict_to_compressed_string(to_dict(metric))


def from_compressed_string(compressed_str: str, project: M.Project) -> M.Metric:
    zipped = base64.b64decode(compressed_str)
    unzipped = zlib.decompress(zipped)
    val = json.loads(unzipped)
    return from_dict(val, project)



from __future__ import annotations

from io import UnsupportedOperation
from typing import Any, Dict, List, Optional, Tuple

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import mitzu.model as M
from mitzu.helper import parse_datetime_input

import mitzu.webapp.pages.explore.dates_selector_handler as DS
import mitzu.webapp.pages.explore.complex_segment_handler as CS
import mitzu.webapp.pages.explore.metric_type_handler as MTH
import mitzu.webapp.pages.explore.toolbar_handler as TH
from dash import html, ctx
import dash_mantine_components as dmc
from mitzu.webapp.helper import (
    CHILDREN,
    METRIC_SEGMENTS,
    find_event_field_def,
    value_to_label,
)

METRICS_CONFIG_CONTAINER = "metrics_config_container"

TIME_WINDOW = "time_window"
TIME_WINDOW_INTERVAL = "time_window_interval"
TIME_WINDOW_INTERVAL_STEPS = "time_window_interval_steps"
AGGREGATION_TYPE = "aggregation_type"
RESOLUTION_DD = "resolution_dd"
RESOLUTION_IG = "resolution_ig"


SUPPORTED_PERCENTILES = [50, 75, 90, 95, 99, 0, 100]


def agg_type_to_str(agg_type: M.AggType, agg_param: Any = None) -> str:
    if agg_type == M.AggType.CONVERSION:
        return "Conversion Rate"
    if agg_type == M.AggType.COUNT_EVENTS:
        return "Event Count"
    if agg_type == M.AggType.RETENTION_RATE:
        return "Retention Rate"
    if agg_type == M.AggType.COUNT_UNIQUE_USERS:
        return "User Count"
    if agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
        return "Average Time To Convert"
    if agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
        if agg_param is None:
            raise ValueError("Time to convert metrics require an argument parameter")
        p_val = round(agg_param)
        if p_val == 50:
            return "Median Time To Convert"
        if p_val == 0:
            return "Min Time To Convert"
        if p_val == 100:
            return "Max Time To Convert"
        return f"P{p_val} Time To Convert"
    raise ValueError(f"Unsupported aggregation type {agg_type}")


def get_time_window_options(metric_type: MTH.MetricType) -> List[Dict[str, int]]:
    res: List[Dict[str, Any]] = []
    if metric_type == MTH.MetricType.CONVERSION:
        for tg in M.TimeGroup:
            if tg in (M.TimeGroup.TOTAL, M.TimeGroup.QUARTER):
                continue
            res.append({"label": tg.name.lower().title(), "value": tg.value})
    elif metric_type == MTH.MetricType.RETENTION:
        for tg in [
            M.TimeGroup.DAY,
            M.TimeGroup.WEEK,
            M.TimeGroup.MONTH,
            M.TimeGroup.YEAR,
        ]:
            if tg in (M.TimeGroup.TOTAL, M.TimeGroup.QUARTER):
                continue
            res.append({"label": tg.name.lower().title(), "value": tg.value})

    return res


def get_agg_type_options(metric_type: MTH.MetricType) -> List[Dict[str, str]]:
    if metric_type == MTH.MetricType.CONVERSION:
        res: List[Dict[str, Any]] = [
            {
                "label": agg_type_to_str(M.AggType.CONVERSION),
                "value": M.AggType.CONVERSION.to_agg_str(),
            },
            {
                "label": agg_type_to_str(M.AggType.AVERAGE_TIME_TO_CONV),
                "value": M.AggType.AVERAGE_TIME_TO_CONV.to_agg_str(),
            },
        ]
        res.extend(
            [
                {
                    "label": agg_type_to_str(M.AggType.PERCENTILE_TIME_TO_CONV, val),
                    "value": M.AggType.PERCENTILE_TIME_TO_CONV.to_agg_str(val),
                }
                for val in SUPPORTED_PERCENTILES
            ]
        )

        return res
    elif metric_type == MTH.MetricType.RETENTION:
        res = [
            {
                "label": agg_type_to_str(M.AggType.RETENTION_RATE),
                "value": M.AggType.RETENTION_RATE.to_agg_str(),
            }
        ]
        return res
    else:
        return [
            {
                "label": agg_type_to_str(M.AggType.COUNT_UNIQUE_USERS),
                "value": M.AggType.COUNT_UNIQUE_USERS.to_agg_str(),
            },
            {
                "label": agg_type_to_str(M.AggType.COUNT_EVENTS),
                "value": M.AggType.COUNT_EVENTS.to_agg_str(),
            },
        ]


def create_metric_options_component(
    metric_config: M.MetricConfig,
    metric_type: MTH.MetricType,
    time_window: M.TimeWindow,
) -> bc.Component:
    if metric_type == MTH.MetricType.SEGMENTATION:
        tw_value = 1
        tw_g_value = M.TimeGroup.DAY
        agg_type = metric_config.agg_type
        agg_param = metric_config.agg_param
        if agg_type not in (M.AggType.COUNT_UNIQUE_USERS, M.AggType.COUNT_EVENTS):
            agg_type = M.AggType.COUNT_UNIQUE_USERS
    elif metric_type == MTH.MetricType.CONVERSION:
        tw_value = time_window.value
        tw_g_value = time_window.period
        agg_type = metric_config.agg_type
        agg_param = metric_config.agg_param
        if agg_type not in (
            M.AggType.PERCENTILE_TIME_TO_CONV,
            M.AggType.AVERAGE_TIME_TO_CONV,
            M.AggType.CONVERSION,
        ):
            agg_type = M.AggType.CONVERSION
    elif metric_type == MTH.MetricType.RETENTION:
        tw_value = time_window.value
        tw_g_value = time_window.period
        if tw_g_value not in [
            M.TimeGroup.DAY,
            M.TimeGroup.WEEK,
            M.TimeGroup.MONTH,
            M.TimeGroup.YEAR,
        ]:
            tw_g_value = M.TimeGroup.WEEK

        agg_type = M.AggType.RETENTION_RATE
        agg_param = None
    else:
        agg_type = M.AggType.COUNT_UNIQUE_USERS
        agg_param = None
        tw_value = 1
        tw_g_value = M.TimeGroup.DAY

    aggregation_comp = dmc.Select(
        id=AGGREGATION_TYPE,
        label="Aggregation",
        className=AGGREGATION_TYPE + " rounded-right",
        clearable=False,
        value=M.AggType.to_agg_str(agg_type, agg_param),
        size="xs",
        data=get_agg_type_options(metric_type),
        style={
            "width": "204px",
        },
    )

    tw_label = (
        "Retention Period" if metric_type == MTH.MetricType.RETENTION else "Within"
    )

    time_window = html.Div(
        [
            dmc.NumberInput(
                id=TIME_WINDOW_INTERVAL,
                label=tw_label,
                className="me-1",
                type="number",
                max=10000,
                min=1,
                value=tw_value,
                size="xs",
                style={"width": "100px", "display": "inline-block"},
            ),
            dmc.Select(
                id=TIME_WINDOW_INTERVAL_STEPS,
                clearable=False,
                value=tw_g_value.value,
                size="xs",
                data=get_time_window_options(metric_type),
                style={"width": "100px", "display": "inline-block"},
            ),
        ],
        style={
            "visibility": "visible"
            if metric_type in [MTH.MetricType.RETENTION, MTH.MetricType.CONVERSION]
            else "hidden"
        },
    )

    resolution_ig = dmc.Select(
        id=RESOLUTION_DD,
        className="rounded-right",
        clearable=False,
        value=metric_config.resolution.name,
        size="xs",
        label="Resolution",
        data=[{"label": value_to_label(v.name), "value": v.name} for v in M.Resolution],
        style={
            "width": "204px",
            "visibility": (
                "visible"
                if metric_type in [MTH.MetricType.RETENTION, MTH.MetricType.CONVERSION]
                else "hidden"
            ),
        },
    )

    return html.Div(children=[aggregation_comp, time_window, resolution_ig])


def from_metric(
    metric_config: M.MetricConfig,
    metric_type: MTH.MetricType,
    time_window: M.TimeWindow,
) -> bc.Component:
    conversion_comps = [
        create_metric_options_component(metric_config, metric_type, time_window)
    ]
    component = dbc.Row(
        [
            dbc.Col(
                children=[DS.from_metric(metric_config, metric_type)],
                xs=12,
                md=6,
            ),
            dbc.Col(children=conversion_comps, xs=12, md=6),
        ],
        id=METRICS_CONFIG_CONTAINER,
        className=METRICS_CONFIG_CONTAINER,
    )
    return component


def from_all_inputs(
    discovered_project: Optional[M.DiscoveredProject],
    all_inputs: Dict[str, Any],
    metric_type: MTH.MetricType,
) -> Tuple[M.MetricConfig, M.TimeWindow]:
    agg_type_val = all_inputs.get(AGGREGATION_TYPE)
    if agg_type_val is None:
        if metric_type == M.MetricType.CONVERSION:
            agg_type, agg_param = M.AggType.CONVERSION, None
        elif metric_type == M.MetricType.SEGMENTATION:
            agg_type, agg_param = M.AggType.COUNT_UNIQUE_USERS, None
        elif metric_type == M.MetricType.RETENTION:
            agg_type, agg_param = M.AggType.RETENTION_RATE, None
        else:
            raise UnsupportedOperation(f"Unsupported Metric Type : {metric_type}")
    else:
        agg_type, agg_param = M.AggType.parse_agg_str(agg_type_val)

    if ctx.triggered_id == MTH.METRIC_TYPE_DROPDOWN:
        start_dt = None
        end_dt = None
        if metric_type == MTH.MetricType.RETENTION:
            time_group = M.TimeGroup.WEEK
            resolution = M.Resolution.ONE_USER_EVENT_PER_HOUR
            lookback_days = M.TimeWindow(2, M.TimeGroup.MONTH)
            time_window = M.TimeWindow(1, M.TimeGroup.WEEK)
        elif metric_type == MTH.MetricType.CONVERSION:
            time_group = M.TimeGroup.DAY
            resolution = M.Resolution.ONE_USER_EVENT_PER_MINUTE
            lookback_days = M.TimeWindow(1, M.TimeGroup.MONTH)
            time_window = M.TimeWindow(1, M.TimeGroup.DAY)
        else:
            time_group = M.TimeGroup.DAY
            resolution = M.Resolution.EVERY_EVENT
            lookback_days = M.TimeWindow(1, M.TimeGroup.MONTH)
            time_window = M.TimeWindow(1, M.TimeGroup.DAY)
    else:
        dates_conf = DS.from_all_inputs(discovered_project, all_inputs)
        time_group = dates_conf.time_group
        resolution = M.Resolution.parse(
            all_inputs.get(RESOLUTION_DD, M.Resolution.EVERY_EVENT.name)
        )
        lookback_days = dates_conf.lookback_days
        start_dt = dates_conf.start_dt
        end_dt = dates_conf.end_dt
        time_window = M.TimeWindow(
            value=all_inputs.get(TIME_WINDOW_INTERVAL, 1),
            period=M.TimeGroup(
                all_inputs.get(TIME_WINDOW_INTERVAL_STEPS, M.TimeGroup.DAY)
            ),
        )

    chart_type_val = all_inputs.get(TH.CHART_TYPE_DD, None)
    chart_type = M.SimpleChartType.parse(chart_type_val)

    group_by: Optional[M.EventFieldDef] = None
    if discovered_project is not None:
        group_by_paths = all_inputs[METRIC_SEGMENTS][CHILDREN]
        if len(group_by_paths) >= 1 and not (
            metric_type == MTH.MetricType.RETENTION and time_group != M.TimeGroup.TOTAL
        ):
            gp = group_by_paths[0].get(CS.COMPLEX_SEGMENT_GROUP_BY)
            group_by = find_event_field_def(gp, discovered_project) if gp else None
            if group_by is not None:
                group_by._project._discovered_project.set_value(discovered_project)

    res_config = M.MetricConfig(
        start_dt=parse_datetime_input(start_dt, None),
        end_dt=parse_datetime_input(end_dt, None),
        lookback_days=lookback_days,
        time_group=time_group,
        agg_type=agg_type,
        agg_param=agg_param,
        chart_type=chart_type,
        resolution=resolution,
        group_by=group_by,
        custom_title="",
    )
    return res_config, time_window
